"""
Optimal Stellar Models (OSM)

Copyright (c) 2012 R. Samadi (LESIA - Observatoire de Paris)

This is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
 
This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this code.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import math
import multiprocessing
import os,sys
import osmlib

class OSMError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value) 

class Parameter:
    name = ''
    value = 0.
    step = 0.
    rate = 0.
    bounds = [0.,0.]
    sigma = -1.
    evol = 0 # = 1 if the parameter controls the evolution
    seismic = 0 # =1 if the parameter control the modes 

    def copy(self):
        new = Parameter()
        new.name = self.name
        new.value = self.value
        new.step = self.step
        new.rate = self.rate
        new.bounds = self.bounds
        new.sigma = self.sigma
        new.evol = self.evol
        new.seismic = self.seismic
        return new

class Target:
    name = ''
    value = 0.
    sigma = -1.
    
    def copy(self):
        new = Target()
        new.name = self.name
        new.value = self.value
        new.sigma = self.sigma
        return new

    
multiproc =  True

def levmar(modelname,func,parameters,func_args,y,covar,f,verbose=1,ftol=1e-3,maxiter=30,chi2min=1e-4,lamb0=1e-4,autostep=False,cov_cdtnb_thr=1e13,hess_cdtnb_thr=1e13):
    '''

	def func (par,args,status): the function that computes the model and that we want to search for the optimal parameters
	
	status : =-1, if the function is called for the central values of the parameters (par)
	         >-1  if the function is called for  the derivatives, in that case status
                  is the index of the parameter for which we cant to compute the derivatives
        parameters : parameters of the function
	args: the other arguments passed to the function (can be 'None' if no others arguments need to be transmitted) 
	
	cov_cdtnb_thr : adopted threshold for the condition number associated with the co-variance matrix
	
	hess_cdtnb_thr : adopted threshold for the condition number associated with the final Hessian matrix
	
	f: output file where results are printed
    
	return : (chi2i,parameters,iter,model,error,msg)
    '''


 
    def Chi2(model,y,W):
        # chi2 =  Chi2(func,par,func_args,y,w )  
        #
        # Y : (IN) le jeux de donnees 
        #
        # model : (IN) le modele 
        #
        # w : (IN)  inverse de la matrice de covariance
        #
        tmp = 0.
        n = len(y)
        for i in range(n):
            for j  in range(n):
                tmp = tmp + (y[i]-model[i]) * W[i,j] * (y[j]-model[j])
        return  tmp

    def levmar_der(model,func, parameters, func_args, y ,  der):
        

        def func_der(parameters,args,index):
            error  = True
            pari = list( i.copy() for i in parameters)
            iter = 0
            while( error == True and iter < 3):
                for i in range(len(pari)):
                    pari[i].value = parameters[i].value
                step = parameters[index].step/(1. + iter)
                pari[index].value = parameters[index].value  + step
                (model,error) = func(pari,func_args, index )
                iter += 1
                if(error):
                    print ('Error in OSM/levmar: unable to compute the derivative (model #%i), we reduce the step by a factor %i') % (index+1,iter+1)
            if(error):    
                print ('Error in OSM/levmar: unable to compute the derivative (mode #%i) after %i iterations') % (index+1,iter)
                model = 0 
            return (model,step,error)
        
        def func_shell(parameters,args,index,pipe):
            error  = True
            (model,error) = func(parameters ,args,  index  )
            pipe.send((model,0.,error))
            pipe.close()
            return

        def func_der_shell(parameters,args,index,pipe):
            error = True
            (model,step,error) = func_der(parameters,args,index)
            pipe.send( (model,step,error)   )
            pipe.close()
            return 

        error = False
        npar = len(parameters)
        models = np.zeros((model.size,npar))
        steps = np.zeros(npar)
        if ( multiproc == True):
            processes=[]
            parents=[]
            childs=[]
            res = []
            pipe = multiprocessing.Pipe()
            parents.append(pipe[0])
            childs.append(pipe[1])
            processes.append(multiprocessing.Process(target=func_shell,args=(parameters,func_args,  -1 , childs[0])))
            processes[0].start()
            nproc = 1
            for i in range(npar):
                if( (parameters[i].evol == 0) & (parameters[i].seismic == 0)):
                    pipe = multiprocessing.Pipe()
                    parents.append(pipe[0])
                    childs.append(pipe[1])
                    processes.append(multiprocessing.Process(target=func_der_shell,args=(parameters,func_args,   i  , childs[nproc])))
                    processes[nproc].start()
                    nproc += 1
            for i in range(nproc):
                try:
                    processes[i].join()
                except:
                    for k in range(nproc):
                        processes[k].terminate()
                    sys.exit(1)

            for i in range(nproc):
                res.append(parents[i].recv())

            error = (res[0])[2]
            if(error):    
                return True
            model[:] = (res[0])[0]
            nproc = 1
            for i in range(npar):
                if( (parameters[i].evol == 0) & (parameters[i].seismic == 0) ):
                    error = (res[nproc])[2]
                    if(error):    
                        return True
                    models[:,i] = (res[nproc])[0]
                    steps[i] = (res[nproc])[1]
                    nproc += 1
                else:
                    (result,step,error) = func_der(parameters,func_args,i)
                    models[:,i] = result
                    steps[i] = step
                    if(error):    
                        return True
        else:
            (result,error) = func(parameters ,func_args, -1 )
            model[:] = result
#            print resutl
            if(error):
                return True
            for i in range(npar):
                (result,step,error) = func_der(parameters,func_args,i)
                models[:,i] = result
                steps[i] = step
#               print result, step
                if(error):    
                    return True

        # on alimente la variable der
        for i in range(npar):
            der[i,:] = ((models[:,i] - model).flatten())/steps[i]

        return error
       
    def levmar_coef(model, npar,  y , W , der, lamb):
        
        alpha=np.zeros((npar,npar))
        beta=np.zeros((npar))
        n = len(y)
        for i in range(npar):
            tmp = 0.
            for l in range(n):
                for m in range(n):
                    tmp = tmp + 0.5 * W[l,m]* ( ( y[l] - model[l]) * der[i,l] +  (y[m] - model[m]) * der[i,m] )
            beta[i] = tmp
            for j in range(i,npar):
                tmp = 0.
                for l in range(n):
                    for m in range(n):
                        tmp = tmp + 0.5 *   W[l,m]* ( der[i,l]*der[j,m]    +  der[i,m]*der[j,l] )
                alpha[i,j] = tmp * (1. + lamb*(i ==j) )
        for i in range(1,npar):
            for j in range(0,i):
                alpha[i,j] = alpha[j,i]
        return (alpha,beta)

    def hessian2str(alpha,parameters):
        text = ("%12s ") % ("")
        npar = len(parameters)
        for p in parameters:
            text += ("%16s ") % (p.name)
        text += '\n'
        for i in range(npar):
            text += ("%12s ") % (parameters[i].name)
            for j in range(npar):
                text += ("%16.10e ") % (alpha[i,j])
            text += '\n'
        return text
        

    text =  '\n~~~~~~~~~~~~~~~~~ OSMlevmar started ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n'
    msg= '' # output message
    
    lamb=lamb0
    text += 'Initial values:\n'
    for p in parameters:
        text += ("%s = %g\n") % (p.name,p.value)
    text +=  ('lambda= %g\n') % (lamb)
    if(verbose):
        print (text)
    f.write(text)
    text = ''
    
    ny  = len(y)
    y =   np.array(y)
    
    # singular value decomposition (SVD): A = U S V
    (U,s,V) = np.linalg.svd(covar)
    #  diagonal matrix that will contain the singular values (eigenvalues)
    Si = np.zeros( (ny,ny) )
    eps = 1e-15
    if( min(np.abs(s)) > 0. ):
        cdtn = max(np.abs(s))/min(np.abs(s)) # condition number
        text += ('Covariance matrix conditioning number: %g\n' ) % (cdtn  )    
    else:
        text2 = 'Error in OSM/levmar: there is a null eigenvalue in the covariance matrix, the covariance matrix cannot be inverted'
        text += text2
        f.write(text)
        f.close()
        raise OSMError(text2)

    # we may need to truncate the matrix 
    limit = max(s)/cov_cdtnb_thr
    truncated = False
    for i in range(ny):
        if(s[i] > limit):
            Si[i,i] =  (1./s[i])  # we retain only eigenvalue s > limit
        else:
            Si[i,i] = 0.
            truncated = True
    if (truncated):
        text2 = '\n\n --> WARNING ! the covariance matrix is truncated using SVD <--\n\n'
        msg  += text2
        text += text2
##    W = np.array(np.linalg.inv(np.matrix(covar)))
    # inverse matrix : A^(-1) = V^t S^(-1) U^t 
    W = np.dot(np.transpose(V),np.dot(Si,np.transpose(U)))
    np.savetxt('covar_inv.txt',W,fmt='%19.8e')

    if(verbose):
        print (text)
    f.write(text)
    text = ''

    
    npar = len(parameters)
    parn=list( p.copy() for p in parameters)
    global der
    der=np.zeros((npar,y.size))
    global dern
    dern=np.zeros((npar,y.size))
    model = np.zeros(ny)
    modeln = np.zeros(ny)
    k=0
    cont=1
    iter=0
    new=True
    error = levmar_der(model,func, parameters , func_args, y ,  der)
    
    if(error):
        text2 = 'Error in OSM/levmar: unable to compute the first model, we exit from levmar'
        text += text2
        if(verbose):
            print text
        f.write(text)
        f.close()
        raise OSMError(text2)

    
    chi2i=Chi2(model,y,W)
    chi2n=chi2i

    def copyfiles(modelname):
        s = modelname + '_fin'
        os.system('\cp -p '+ modelname +'_B.dat ' + s+'_B.dat')
        os.system('\cp -p  '+ modelname +'.don ' + s+'.don')
        os.system('\cp -p  '+ modelname +'.HR ' + s+'.HR')
        if ( os.access(modelname+'-nad.osc',os.F_OK)  ):
            os.system('\cp -p  '+ modelname +'-nad.osc ' + s+'-nad.osc')
        else:
            os.system('\cp -p  '+ modelname +'-ad.osc ' + s+'-ad.osc')
        if ( os.access(modelname+'.gsm',os.F_OK)  ):
            os.system('\cp -p  '+ modelname +'.gsm ' + s+'.gsm')

    copyfiles(modelname)

    text += '\n##############################################\n'
    text += "Initial model:\n"
    text += ('chi2= %g\n') % (chi2i)
    text += 'Parameters:\n'
    for p in parameters:
            text += ("%s = %g\n") % (p.name, p.value)
    text += "\n"
    text += "Model:" +  np.array_str(model) + "\n"
    text += "Constraints: " +  np.array_str(y) + "\n"
    #    text += "Sigma: " +  np.array_str(sigma) + "\n"
    text += "Distances (target-model):" +  np.array_str(y - model) + "\n"

    if(verbose):
        print (text)
    f.write(text)
    f.flush()
    modeln[:] = model[:]
    def sign(a):
        if(a < 0):
            return -1.
        else:
            return 1.
        
    while(cont):
        text =  '\n##############################################\n\n'
        text += ('Iter # %i\n') % (iter)

        (alpha,beta)=levmar_coef(model,npar,y,W,der,lamb)
        text += "Hessian matrix:\n"
        text += hessian2str(alpha,parn)
        # singular value decomposition (SVD): A = U S V
        (U,s,V) = np.linalg.svd(alpha)
        if( min(np.abs(s)) > eps ):
            cdtn = max(np.abs(s))/min(np.abs(s)) # condition number
            text += ('Hessian matrix conditioning number: %g\n' ) % (cdtn)    
        else:
            text2 = 'Error in OSM/levmar: there is a null eigenvalue in the Hessian matrix, it cannot be inverted'
            text += text2
            f.write(text)
            f.close()
            raise OSMError(text2)

## 	# we may need to truncate the matrix 
## 	limit = max(s)*1e-13
## 	truncated = False
## 	#  diagonal matrix that will contain the singular values (eigenvalues)
## 	Si = np.zeros( (npar,npar) )
## 	for i in range(npar):
## 		if(s[i] > limit):
## 			Si[i,i] =  (1./s[i])  # we retain only eigenvalue s > limit
## 		else:
## 			Si[i,i] = 0.
## 			truncated = True
## 	if (truncated):
## 		text2 =  '\n\n --> WARNING ! the Hessian matrix is truncated using SVD <--\n\n'
## 		text += text2
## 	# inverse matrix: A^(-1) = V^t S^(-1) U^t 
## 	alpha = np.dot(np.transpose(V),np.dot(Si,np.transpose(U)))
        alpha = np.linalg.inv(np.matrix(alpha))

        dpar= (np.array(alpha * np.matrix(beta.reshape((npar,1))))).flatten()
        for i in range(npar):
            if( math.fabs(dpar[i]) >  math.fabs(parameters[i].value*parameters[i].rate/100.) ):
                dpar[i] = math.fabs(parameters[i].value*parameters[i].rate/100.) * sign(dpar[i])
            if( ( parameters[i].value+ dpar[i] - parameters[i].bounds[0]   < 0 ) ):
                dpar[i] = parameters[i].bounds[0] - parameters[i].value 
            elif( (  parameters[i].value+ dpar[i] - parameters[i].bounds[1]  > 0 ) ):
                dpar[i] = parameters[i].bounds[1] - parameters[i].value
            parn[i].value  = parameters[i].value+ dpar[i]
            if (autostep):
                # updating the step used for the calculation of the function derivatives
                refstep = parameters[i].step
                newstep = 3.*math.sqrt(math.fabs(alpha[i,i]))
    ##            newstep = 3.*math.fabs(dpar[i])
                parn[i].step = max( min(newstep, refstep*1e1),refstep*1e-1)
    ##            print ("refstep, nstep,dpar, i = %f,  %f , %f , %i") %(refstep,parn[i].step,dpar[i] , i)
        

        text += 'New parameters: name = value (step)\n'
        for p in parn:
            text += ("%s = %g (%f)\n") % (p.name,p.value,p.step)
        text += 'Change (new-old): ' +  np.array_str(dpar) + "\n"
        text +=  '.................................................\n'

        if(verbose):
            print text
        f.write(text)
        f.flush()

        error = levmar_der(modeln,func,parn, func_args, y , dern)
        if(error):
            text = ('Error in OSM/levmar: unable to compute the model at iteration # %i, we exit from levmar')  % (iter)
            if(verbose):
                print text
            f.write(text)
            f.close()
            raise OSMError(text)

        chi2n=Chi2(modeln,y,W)
        text = ("New chi2: %g\n") % (chi2n)
        text += ("Best chi2: %g\n") % (chi2i)       
        text += "Parameters: "
        for p in parn:
            text += ("%g ") % (p.value)
        text += "\n"
        text += "Model:" +  np.array_str(modeln) + "\n"
        text +=  "Constraints: " +  np.array_str(y) + "\n" 
        #        text +=  "Sigma: " +  np.array_str(sigma) + "\n" 
        text +=  "Distances (target-model):" +  np.array_str(y - modeln) + "\n"

        df = (chi2n-chi2i)/chi2i # relative variation of the chi2
        cont= (iter < maxiter) & (abs(df) > ftol ) & (chi2n > chi2min)
        if(chi2n>=chi2i):
            lamb = lamb*10.
            text += 'Leaving the parameters unchanged: \n'
        else:
            lamb = lamb/4.
            for i in range(npar):
                parameters[i].value =parn[i].value
            chi2i=chi2n
            model[:]=modeln[:]
            der[:,:]=dern[:,:]
            text += 'Adopted parameters: \n' 
            for p in parameters:
                text += ("%s = %g\n") % (p.name,p.value)

            copyfiles(modelname)

        text +=  ('chi2 , dchi2/chi2 , lambda : %g %g %g\n') % (chi2i ,  df , lamb)
        text +=  '\n\n'
        if(verbose):
            print text
        f.write(text)
        f.flush()
        iter=iter+1

    text =  '\n##############################################\n'
    
    text2 = "Stopped because:\n"
    if(iter >= maxiter):
        text2 += ("\tnumber of iterations >= %i\n") % (maxiter)
    if(abs(df) <= ftol ):
        text2 += ("\trelative variation of the chi2 <  %g\n") % (ftol)
    if( chi2n <= chi2min):
        text2 += ("\tchi2 <= %g\n") % (chi2min)
    msg += text2
    text += text2
    text += ('chi2 = %g\n') % (chi2i)
    text += ('lambda = %g\n') % (lamb)
    text += ('dchi2/chi2 = %g\n') % (df)
    text += ('number of iteration = %i\n\n') % (iter)
    text +=  ("Reduced Chi2 = %g\n\n") %  (chi2i/float(ny))

    lamb = 0.
    (alpha,beta)=levmar_coef(model,npar,y,W,der,lamb)
    text += 'Hessian matrix:\n'
    text += hessian2str(alpha,parameters)
    np.savetxt('hessian.txt',alpha,fmt='%19.8e')
    
    # singular value decomposition (SVD): A = U S V
    (U,s,V) = np.linalg.svd(alpha)
    if( min(np.abs(s)) > eps ):
        cdtn = max(np.abs(s))/min(np.abs(s)) # condition number
        text += ('Final Hessian matrix conditioning number: %g\n' ) % (cdtn)    
    else:
        text2 = 'Error in OSM/levmar: there is a null eigenvalue in the Hessian matrix, it cannot be inverted'
        text += text2
        f.write(text)
        f.close()
        raise OSMError(text2)
    # we may need to truncate the matrix 
    limit = max(s)/hess_cdtnb_thr
    truncated = False
    #  diagonal matrix that will contain the singular values (eigenvalues)
    Si = np.zeros( (npar,npar) )
    for i in range(npar):
        if(s[i] > limit):
            Si[i,i] =  (1./s[i])  # we retain only eigenvalue s > limit
        else:
            Si[i,i] = 0.
            truncated = True
    if (truncated):
        text2 =  '\n\n --> WARNING ! the final Hessian matrix is truncated using SVD <--\n\n'
        msg  += text2
        text += text2
    
##    W = np.array(np.linalg.inv(np.matrix(alpha)))
    # inverse matrix: A^(-1) = V^t S^(-1) U^t 
    alpha = np.dot(np.transpose(V),np.dot(Si,np.transpose(U)))
    np.savetxt('hessian_inv.txt',alpha,fmt='%19.8e')
    for i in range(npar):
        parameters[i].sigma = math.sqrt(alpha[i,i])
        
    text += '\nFinal values:\n'
    fparam = open('param.txt','w')
    fparam.write("# parameter: name value sigma\n")
    for p in parameters:
        text += ("%s = %g  +/- %g\n") % (p.name,p.value,p.sigma)
        fparam.write( ('%s %g %g\n') % (p.name,p.value,p.sigma) )
    text +=  '\n'
    fparam.close()
    '''  
    for s in y:
	text += ("%g ") % (s)
    for i in range(ny):
        text += ("%g ") % (math.sqrt(covar[i,i]))
    for i in range(len(model)):
	text += ("%g ") % (model[i])
    for s in parameters:
    	text += ("%g ") % (s.value)
    for s in parameters:
        text += ("%g ") % (s.sigma)
    text += ("%g\n") % (chi2i)
    '''
    
    text +=  '\n~~~~~~~~~~~~~~~~~ OSMlevmar finished ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n'

    if(verbose):
        print (text)

    f.write(text)
    return (chi2i,parameters,iter,model,error,msg)




#!/usr/bin/python

"""
Optimal Stellar Models (OSM)

Copyright (c) 2012 R. Samadi (LESIA - Observatoire de Paris)

This is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
 
This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this code.  If not, see <http://www.gnu.org/licenses/>.
"""
 
__version__ = "1.4"

import os,sys,traceback
import getopt
from osmlib import *
import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.MIMEBase import MIMEBase
from email import Encoders

def search_guess():
    
    ans = ''
    gridname = 'grid'
    path = '/data/samadi/models/grids/mlt/cestam/HR/m00/'
    ans = (raw_input("The grid name: " + gridname))
    if(ans != ''):
        gridname = ans
    ans = (raw_input("Full path: " + path))
    if(ans != ''):
        path = ans
    while( ans != 'q'):
        print "Type 'q' to quit"
        ans = (raw_input("log g: ")).lower()
        if(ans != 'q'):
            logg = float(ans)
            ans = (raw_input("Teff: ")).lower()
            if(ans != 'q'):
                teff = float(ans)
                search_model_in_grid(logg,teff,gridname,path,dteff=50., dlogg=0.1,old=False)           
    

def main():

    def copyright():

        print    "\n\
        Optimal Stellar Models (OSM)\n\
        \n\
        Copyright (c) 2012 R. Samadi (LESIA - Observatoire de Paris)\n\
        \n\
        This is a free software: you can redistribute it and/or modify\n\
        it under the terms of the GNU General Public License as published by\n\
        the Free Software Foundation, either version 3 of the License, or\n\
        (at your option) any later version.\n\
        \n\
        This software is distributed in the hope that it will be useful,\n\
        but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
        GNU General Public License for more details.\n\
        You should have received a copy of the GNU General Public License\n\
        along with this code.  If not, see <http://www.gnu.org/licenses/>.\n"


    def usage():
        print "This is OSM version: " + __version__
        print "usage : osm [-e] [-r] name.xml "
        print "usage : osm -G"
        print "usage : osm -v"

        print "Option:"
        print "-e <address> : send an email to <address> after completion"
        print "-G : search guess parameters in a grid"
        print "-v : print code version and exit"
        print "-r : starts from previous parameter values, \
        the values are take from the directory containing the previous successful calculation "

    if(len(sys.argv)<2):
        copyright()
        usage()
        sys.exit(2)
    try:
        opts,args = getopt.getopt(sys.argv[1:],"he:G:vr")

    except getopt.GetoptError, err:
        print str(err)
        usage()
        sys.exit(2)
        
    email = ''
    guess_flag = False
    resume = False
    for o, a in opts:
        if o == "-h" :
            usage()
            sys.exit(1)
        elif o == "-v":
            print __version__
            sys.exit(1)
        elif o == "-e":
            email = a.strip()
        elif o == "-G" :
            guess_flag = True
        elif o == "-r":
            resume = True
        else:
            print "unhandled option %s" % (o)
            sys.exit(1)
             
    print "This is OSM version: " + __version__

    if(guess_flag):
        search_guess()
        sys.exit(1)

    nargs = len(args)
    if nargs > 1 :
        print "too many arguments"
        usage()
        sys.exit()

    if nargs < 1 :
        print "missing arguments"
        usage()
        sys.exit()


    name =  re.sub('\.xml','',os.path.basename(args[0]))

    if(email != ''):
        sys.stdout = open(name+'.log', 'w')

    copyright()

    cwd = os.getcwd()

    try:
        parameters,targets,settings,seismic_constraints = osminit(name,resume=resume)
    except OSMError as e:
        results=  traceback.format_exc()
        raise
    except:
        results  = traceback.format_exc()  # ("Unexpected error: %s") % (sys.exc_info()[0])
        raise
    
    if(email != ''):
        print ''
        print 'Running OSM, results will be sent to ' + email
        sys.stdout.flush()
        # Create the container (outer) email message.
        msg = MIMEMultipart()
        s = ('OSM: starting %s (job PID= %d)')  % ( name ,  os.getpid()  )
        msg['Subject'] = s
        msg['From'] = email
        msg['To'] = email
        msg.preamble = 'OSM'
        msg.attach( MIMEText(s))
        
        f = name+'.log'
        part = MIMEBase('text', "plain")
        part.set_payload( open(f,"rb").read() )
        Encoders.encode_base64(part)
        part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(f))
        msg.attach(part)       

        f = name+'.xml'
        part = MIMEBase('text', "xml")
        part.set_payload( open(f,"rb").read() )
        Encoders.encode_quopri(part)
        part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(f))
        msg.attach(part)
        
        f = name+'.don'
        part = MIMEBase('text', "plain")
        part.set_payload( open(f,"rb").read() )
        Encoders.encode_base64(part)
        part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(f))
        msg.attach(part)       

        # Send the email via our own SMTP server.
        try:
            s = smtplib.SMTP('localhost')
        except: # Send the email via smtp.obspm.fr
            try:
                s = smtplib.SMTP('smtp.obspm.fr')
            except:
                sys.stdout.close()
                sys.stdout = sys.__stdout__
                print 'an error occurs while sending the email'
                raise

        try:
            s.sendmail(email, email, msg.as_string())
            s.quit()
        except:
            sys.stdout.close()
            sys.stdout = sys.__stdout__
            print 'an error occurs while sending the email'
            raise

    if(email != ''):
        sys.stderr = open(name+'.err', 'w')
       
    try:
        results = osmrun(name,parameters,targets,settings,seismic_constraints,resume=resume)
    except OSMError as e:
        results=  traceback.format_exc()
        if(email == ''):
            print results
            raise
    except:
        results  = traceback.format_exc()  # ("Unexpected error: %s") % (sys.exc_info()[0])
        if(email == ''):
            print results
            raise

    os.chdir(cwd)
    sys.stdout.flush()  
    sys.stderr.flush()
    if(email != ''):
        # Create the container (outer) email message.
        msg = MIMEMultipart()
        msg['Subject'] = 'OSM: results for  ' + name
        msg['From'] = email
        msg['To'] = email
        msg.preamble = 'OSM'

        msg.attach( MIMEText(results ))

        f = name+'.log'
        part = MIMEBase('text', "plain")
        part.set_payload( open(f,"rb").read() )
        Encoders.encode_base64(part)
        part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(f))
        msg.attach(part)       

        # Send the email via our own SMTP server.
        try:
            s = smtplib.SMTP('localhost')
        except: # Send the email via smtp.obspm.fr
            try:
                s = smtplib.SMTP('smtp.obspm.fr')
            except:
                raise
        try:
            s.sendmail(email, email, msg.as_string())
            s.quit()
        except:
            raise
        

if __name__ == "__main__":
    main()




"""
Optimal Stellar Models (OSM)

Copyright (c) 2012 R. Samadi (LESIA - Observatoire de Paris)

This is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
 
This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this code.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import os,sys
import scipy.interpolate 
import math
import cestam
import xml.dom.minidom as xdm
import re
import cestam.constants
from osmlevmar import *
import matplotlib.pyplot as plt
from osmmodes import *
import struct
from scipy import interpolate

class Target:
    name = ''
    value = 0.
    sigma = -1.
    
    def copy(self):
        new = Target()
        new.name = self.name
        new.value = self.value
        new.sigma = self.sigma
        return new

class OSMError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value) 

    
def files_search(dirs,path='./',pattern=''):

    test = re.compile(pattern, re.IGNORECASE) 


    files = []
    for dir in dirs: 
        for root, dirnames, filenames in os.walk(os.path.normpath(path)+'/'+dir):
            f = sorted(filenames)
            for filename in f:
                if(re.search(test,filename)):
                    files.append(root+'/'+filename)
    return files


def loadhrgrid(files,ntmax=10000,old=False):
    '''
 files : names of the files (full path)

  return a tuple (data,masses,ntimes)
 where data  is a  Numpy array data[k,j,i]:
  k : track index (track associated with files[k])
  j : index associated with the following quantities:
    0 : age (My)    
    1 : log Teff
    2 : log L/Lo
    3 : log R/Ro
    4 : M/Mo
    5 : log g
    6 : Xc
    7 : Yc
    8 : Zc
    9 : Xs
   10 : Ys
   11 : Zs
  i : time step index (age)
  Note that age[k] = -1 for  i>ntimes[k]
  masses[k] : masse associated with track k
 the number of time steps associated with trak k
    '''
    
    nfiles = len(files)
    data = np.zeros((nfiles,12,ntmax))
    masses = np.zeros(nfiles)
    ntimes = np.zeros(nfiles)
    cwd = os.getcwd()
    k = 0
    old = False
    for f in files:
        file =  re.sub('\.HR','',os.path.basename(f))
        
        os.chdir(os.path.dirname(f))
        print file
        m = cestam.CModel(file, read=False)
        try:
            m.read_hr(old=old)
        except:
            try:    
                m.read_hr(old=True)
            except:
                k += 1
                print 'in loadhrgrid : unable to read ' + file + ', skipping this file'
                continue
                old = True
        nt = len(m) 
        masses[k] = m.mstar[0]
        ntimes[k] = nt
        if(nt < ntmax):
            data[k,0,nt:] = -1.
        else:
            nt = ntmax
        data[k,0,0:nt] = m.age[0:nt]
        data[k,1,0:nt] = m.log_teff[0:nt]
        data[k,2,0:nt] = m.log_l[0:nt]
        data[k,3,0:nt] = m.log_r[0:nt]
        data[k,4,0:nt] = m.mstar[0:nt]
        data[k,5,0:nt] = m.log_g[0:nt]

        data[k,6,0:nt] = m.ab_c[0,0:nt] # Xc
        data[k,7,0:nt] = m.ab_c[1,0:nt] + m.ab_c[2,0:nt] # Yc
        data[k,8,0:nt] = (m.ab_c[:,0:nt]).sum(0) # Zc

        data[k,9,0:nt] = m.ab_s[0,0:nt] # Xs
        data[k,10,0:nt] = m.ab_s[1,0:nt] + m.ab_s[2,0:nt] # Ys
        data[k,11,0:nt] = (m.ab_s[:,0:nt]).sum(0) # Zs
        
        k += 1
     
    os.chdir(cwd)
    return (data,masses,ntimes)
   


def search_model(hrin,masses,ntimes,logg,teff, dlogg=0.01,dteff=15. ,interactif=False):

    print 'teff = ', teff
    print 'log g =',logg
    logteff = math.log10(teff)

    vicinity = 0.05

    logg_sun = 4.438

    hr = hrin.copy()

    nmasses = masses.size
    ntmax = hr.shape[2]
    (massi,ri,agei) = (0.,0.,0.)
    ans = ''
    while( ans != 'n'):
        if(interactif):
            print  '\nChoose the evolutionnary status:'
            print 'Any -> a'
            print 'Main-sequence -> m'
            print 'Post-sequence -> p'
            print 'Helium burning phase -> h'
            print ''
            print 'Type n to exit\n'
            choices = ['a','m','p','n']
            ans = ''
            while ans not in choices:
                ans = raw_input("Your choice: ")
                ans = ans.lower()
        else:
            ans = 'a'
        selected = []
        if(ans != 'n'):
            
            for k in range(nmasses):
                if(ans == 'm'):
                    j = np.where(hr[k,6,0:ntimes[k]] > 1e-2)
                    if(len(j) > 0):
                        selected.append(j[0])
                elif (ans == 'p'):
                    j = np.where(hr[k,6,0:ntimes[k]] <= 1e-2)
                    if(len(j) > 0):
                        selected.append(j[0])
                elif (ans == 'h'):
                    j = np.where(hr[k,7,0:ntimes[k]] >= 1e-5)
                    if(len(j) > 0):
                        selected.append(j[0])
                else:
                    selected.append(np.arange(ntimes[k],dtype=np.int))  
            if(len(selected) ==0):
                print 'No models in the selected evolutionnary status'
        if( (ans != 'n') & (len(selected) >0)):
            chi2 = np.ones((nmasses,ntmax))*1e10
            distance = np.ones((nmasses,ntmax))*1e10

            for k in range(nmasses):
                if(len(selected[k]) > 0):
                    # ecarts quadratique:
                    chi2[k,selected[k]] = ((10.**hr[k,1,selected[k]]-teff)/dteff )**2 +  ((hr[k,5,selected[k]]-logg)/dlogg)**2
                    # distance eulerienne:
                    distance[k,selected[k]] = np.sqrt((hr[k,1,selected[k]]-logteff)**2 +  (hr[k,5,selected[k]]-logg)**2)

            # on recherche le minimum
            chi2min = 1e10
            distancemin = 1e10

            imin = -1
            kmin = -1
            for i in range(nmasses):
                k = np.argmin( distance[i,:] )
                minimun = distance[i,k]

                if( minimun  < distancemin):
                    distancemin  = minimun
                    imin = i
                    kmin = k

            chi2min = chi2[imin,kmin]

            plt.figure(0)
            plt.clf()
            for k in range(nmasses):
                if(len(selected[k])>0):
                    plt.plot(hr[k,1,selected[k]],hr[k,5,selected[k]])
            plt.plot([logteff],[logg],'bs')
            plt.axis([3.9,3.5,5,1])

            if( imin < 0 or kmin < 0):
                return (0.,0.)

            #  modele correspondant au minimum du chi2:
            print 'closest model:'
            print 'chi2=',chi2min
            print 'distance=',distancemin
            print 'mass=', masses[imin]
            print 'age=', hr[imin,0,kmin]
            teff0 = 10.**hr[imin,1,kmin]
            logteff0 = math.log10(teff0)
            print 'teff=',  teff0
            print 'log teff=',  logteff0
            print 'r=', 10.**hr[imin,3,kmin]
            logg0 = hr[imin,5,kmin]
            print 'logg=', logg0

            plt.figure(1)
            plt.clf()
            for k in range(nmasses):
                if(len(selected[k])>0):
                    plt.plot(hr[k,1,selected[k]],hr[k,5,selected[k]])
            logg_v = np.array([])
            logteff_v = np.array([])
            r_v = np.array([])
            age_v = np.array([])

            for i in range(nmasses-1):
                if(len(selected[i]) > 0) :
                    j = np.where( (hr[i,1,selected[i]]  >= logteff-vicinity)  & (hr[i,1,selected[i]]  <= logteff+vicinity) &  (hr[i,5,selected[i]]  >= logg-vicinity)  & (hr[i,5,selected[i]]  <= logg+vicinity ) ) 
                    if (j[0].size >0):
                        u = (selected[i])[j[0]]
                        logg_v = np.append(logg_v,hr[i,5,u])
                        logteff_v= np.append(logteff_v,hr[i,1,u])
                        r_v = np.append(r_v,10.**hr[i,3,u])
                        age_v = np.append(age_v,hr[i,0,u])
            n= r_v.size

            plt.plot(logteff_v,logg_v,'go')

            plt.plot([logteff0],[logg0],'r^')
            plt.plot([logteff],[logg],'bs')

            plt.axis([logteff+vicinity,logteff-vicinity,logg+vicinity,logg-vicinity ])

            ri = float(scipy.interpolate.griddata( (logg_v,logteff_v) , r_v , (logg,logteff) , method='cubic'))
            agei = float(scipy.interpolate.griddata( (logg_v,logteff_v) , age_v , (logg,logteff) , method='cubic'))

            print 'Interpolated radius Ri=', ri

            massi = 10**(logg-logg_sun)*ri**2
            print 'Mass associated with Ri and log g:', massi
            print 'Interpolated age: ', agei

            plt.show()

    return  (massi,ri,agei)

def search_model_in_grid(logg,teff,gridname,path,dteff=50., dlogg=0.1,old=False):
    
    if (  not(os.access(path + gridname + '.npz',os.F_OK))):
        pattern='.HR'
        #        dirs= ['m0.8_m1.18','m1.2_m1.98','m2.0_3.95','m4.0_5.95']
        dirs = ['.']
        files= files_search(dirs,path=path,pattern='.HR')
        (hr,masses,ntimes) = loadhrgrid(files,old=old)
        np.savez(path+gridname,hr=hr,masses=masses,ntimes=ntimes)
    else:
        print 'Loading  ' + gridname + '.npz'
        hrdata = np.load(path + gridname +'.npz')
        hr = hrdata['hr']
        masses = hrdata['masses']
        ntimes = hrdata['ntimes']
    
    search_model(hr,masses,ntimes,logg,teff,dteff=dteff, dlogg = dlogg ,interactif=True)

sigma_stefan = cestam.constants.clight * cestam.constants.aradia/4.
logg_sun = math.log10(cestam.constants.ggrav) + math.log10(cestam.constants.msun) - 2.*math.log10(cestam.constants.rsun)

def get_first_value(name,dom) :
    item = dom.getElementsByTagName(name)
    if ( len(item) > 0):
        return item[0].childNodes[0].data
    else:
        None

def clean_string(string):
    return re.sub("\s{1,}$","",re.sub("^\s{1,}","",string))

def write_amdl(file,data,AA, record_maker = 4, endian = '@'):
    '''
 INPUTS:
 file : name of the output file
 data : global parameters  (see ADIPLS user's manual)
 AA   : variables (see ADIPLS user's manual)
  AA[0,:] = x
  AA[1,:] = A1
  ...
  AA[n,:] = An

 OPTIONS:
 endian (='@'): to be setup depending on your system architecture, see the module 'struct'
 record_maker (=4) :  to be setup depending on the way adipls is compiled, look at the makefile and the compiler options

 Return :   0, if successful, 1 otherwise
    '''
    
    if(record_maker == 4): # 32 bits
        HEADER_PREC = 'i'
    elif(record_maker  == 8): # 64 bits
        HEADER_PREC = 'l'
    else:
        print 'Error in write_amdl: value of record_marke  not handled'
        return 1

    if( record_maker > struct.calcsize("P") ):
        print 'Error in write_amdl: record_marker not compatible with your architecture'
        return 1
      
    f = open(file,'w')
    n = AA.shape[1]
    nb = AA.shape[0]*n*8 + len(data)*8  + 2*4  # 2 x 4-bytes integers
    if(AA[0,0] > AA[0,n-1]):
        B = ((AA[:,::-1]).transpose()).flatten()
    else:
        B = (AA.transpose()).flatten()
    f.write(struct.pack(endian+HEADER_PREC,nb))
    f.write(struct.pack(endian+'ii',1,n)) # 2x 4-bytes interger
    for s in data:
        f.write(struct.pack(endian+'d',s))
    for s in B:
        f.write(struct.pack(endian+'d',s))
    f.close()

    return 0


def cesam2amdl(model,bv=0):
    """
    (data,AA) = cesam2amdl(model,bv=0)
    
    INPUTS:
    model : cestam model
    bv  : if bv=1 the Brunt-Vaisala freq. is computed numerically, otherwise it is taken from the cestam model

    OUTPUT:
    ADIPLS model
    
    """
    nmodel = model.ntot
    data  = np.zeros(8)
    data[0] = model.glob[0] # M
    data[1] = model.glob[1] # R
    data[2] = model.var[3,nmodel-1]  #  Pc
    data[3] = model.var[4,nmodel-1]  #  rhoc
    data[4] = -model.glob[8]/model.var[9,nmodel-1] # -(d2p/dx^2)/p/Gamma1 at the center
    data[5] = -model.glob[9] # -(d2dho/dx^2)/rho at the center
    data[6] = -1.
    data[7] = 0.
    AA = np.zeros( (6,nmodel))
    # cesam.ggrav = 6.672320e-08
    
    AA[0,0:nmodel-1] = model.var[0,0:nmodel-1]/model.glob[1] # x = r/R
    AA[1,0:nmodel-1] = model.var[1,0:nmodel-1]/(AA[0,0:nmodel-1])**3 # (m/M)/x^3
    AA[2,0:nmodel-1] = cestam.ggrav*(model.var[1,0:nmodel-1]*model.glob[0]) * model.var[4,0:nmodel-1]/(model.var[9,0:nmodel-1]*model.var[3,0:nmodel-1]*model.var[0,0:nmodel-1])  # G*m*rho/ (Gamma1*p*r)
    AA[3,0:nmodel-1] = model.var[9,0:nmodel-1] # Gamma1
    if (bv !=0):
        # compute numerically brunt-vaisala freq.
        y = np.log(model.var[4,:-1])
        x = np.log(model.var[0,:-1])
        tck = interpolate.splrep(x[::-1],y[::-1],s=0)
        AA[4,0:nmodel-1] = - AA[2,0:nmodel-1] - interpolate.splev(x,tck,der=1)
    else:        
        AA[4,0:nmodel-1] = model.var[14,0:nmodel-1] # vaisala
    AA[5,0:nmodel-1] = 4.*math.pi*model.var[4,0:nmodel-1]* model.var[0,0:nmodel-1]**3 / (model.glob[0]*model.var[1,0:nmodel-1]) # U = 4pi*rho*r^3/m

    # at the center, asymptotic values:
    AA[0,nmodel-1] = 0.
    AA[1,nmodel-1] = 4.*math.pi*data[1]**3*data[3]/(3.*data[0])
    AA[2,nmodel-1] = 0.
    AA[3,nmodel-1] = model.var[9,nmodel-1] # Gamma1
    AA[4,nmodel-1] = 0.
    AA[5,nmodel-1] = 3.

# data must be filled from the center up to the surface
    if(AA[0,0] > AA[0,nmodel-1]):
        AA = AA[:,::-1]

    return (data,AA)
   
    
def pmsmodel(model,mass):
    print ('\nWe build a PMS model with mass = %g') % (mass)

    model.GUI = False
    model.run.type_file = 'ASCII'
    model.params.agemax = 0.
    model.params.mtot = mass
    
##     X0 = model.params.x0
##     if(X0 < 0):
##         X0 = (1. - model.params.y0)/(1. + model.params.zsx0)
##     model.params.x_stop = X0*(1. - 1e-4)
##     model.params.arret= 'Other'
    
    if(model.run.c_iben < 8.1e-5):
        mod_init = '8d-5.pms'
    elif (model.run.c_iben < 5.1e-4):
        mod_init = '5d-4.pms'
    else:
        mod_init = '2d-2.pms'

    if( not(os.access(model.params.nom_chemin + '/' + mod_init,os.F_OK)))  :
        raise OSMError("Error in OSM: file " + model.params.nom_chemin + '/' + mod_init+" is missing")

    model.run.job = 'From PMS'
    model.run.mod_init = model.params.nom_chemin + mod_init
    model(mkdon=False)
    model.run.type_file = 'Binary'

def zamsmodel(model,mass):
    print ('\nWe build a ZAMS model with mass = %g') % (mass)

    model.GUI = False
    model.run.type_file = 'ASCII'
    model.params.agemax = 0.
    
##     X0 = model.params.x0
##     if(X0 < 0):
##         X0 = (1. - model.params.y0)/(1. + model.params.zsx0)
##     model.params.x_stop = X0*(1. - 1e-4)
##     model.params.arret= 'Other'
    
    if(mass < 1.5):
        mod_init = 'm010.zams'
        mass0 = 1.
    elif (mass < 3.5):
        mod_init = 'm020.zams'
        mass0 = 2.
    else:
        mod_init = 'm050.zams'
        mass0 = 5.

    dmass = mass*0.02
    if(mass<mass0):
        dmass = - dmass

    if( not(os.access(model.params.nom_chemin + '/' + mod_init,os.F_OK)))  :
        raise OSMError("Error in OSM: file " + model.params.nom_chemin + '/' + mod_init+" is missing")

    print ('\nstarting with the initial m = %g') % (mass0)
    print ('proceeding then iteratively with a step dm=%g') % (dmass)

    model.params.mtot = mass0
    model.run.job = 'From ZAMS'
    model.run.mod_init = model.params.nom_chemin + mod_init
    model(mkdon=False)
    model.run.type_file = 'Binary'
    model.run.mod_init = model.name + '_B.hom'
    model(mkdon=False)
    while (model.finished):
        model.params.mtot  = model.params.mtot + dmass
        if( abs(model.params.mtot - mass) < abs(dmass)*1.001 ):
            break
        print ('\ncomputing m=%g') % (model.params.mtot)
        model.run.mod_init = model.name + '_B.hom'	
        model()
        
    model.params.mtot = mass
    print ('\ncomputing the final value m=%g') % (mass)

    model.run.mod_init = model.name + '_B.hom'	
    model(mkdon=False)

def setupparams(Model,Params,Settings):
    for s in Params:
        found = False
        for a in Model.params.nlist.values():
                for b in a:
                        if(b == s.name):
                                found = True
                                Model.params.__setattr__(b,s.value)
        if ( (s.name == 'se_a') | (s.name == 'se_b') | (s.name == 'se_c') ): # parametes used for the surface effects ('se')
            found = True
        if (found):
            print ("%s = %g") % (s.name,s.value)
        if(not(found)):
            raise OSMError("Error in OSM/setupparams: Parameter named "+  s.name + " not found")
    # managing the diffusion:
    if( Model.params.diffusion ):
        zsx0 = -1.
        y0 = -1.
        for s in Params:
            if( s.name == 'zsx0' ):
                zsx0 = Model.params.zsx0
            if( s.name == 'y0' ):
                y0 = Model.params.y0
        if(  zsx0*y0 < 0. ):
            a = Settings['models']['dy_dz']
            yp = Settings['models']['yp']
            zp = Settings['models']['zp']
            if(y0 <0.):
                y0 = 1. - (1.+zsx0)*(1.+a*zp-yp)/(1.+zsx0+a*zsx0)
            if(zsx0<0.):
                zsx0 = (yp-y0-a*zp)/(y0*(1.+a)-a+a*zp)
            print 'dy_dz,yp,zp' , a , yp, zp

        Model.params.y0 = y0
        Model.params.zsx0 = zsx0
        Model.params.x0 = (1. - Model.params.y0)/(1. + Model.params.zsx0)
        Model.params.z0 = (1. - Model.params.x0 - Model.params.y0)
    # managing the diffusion:
    else:          
        if( Model.params.x0<0):
            Model.params.x0 = (1. - Model.params.y0)/(1. + Model.params.zsx0)
            Model.params.z0 = (1. - Model.params.x0 - Model.params.y0)

    print 'x0,y0,z0,zsx0 = ',Model.params.x0, Model.params.y0,Model.params.z0,Model.params.zsx0
        
def compute_optimal(modelname,Parameters,Targets,Settings,SeismicConstraints,Aux3DData=None,resume=True):


    def getoutputs(Model,Params):
        if(Aux3DData is not None):
            pb = Aux3DData['pb'] # pressure at bottom (it is the total pressure if the tag pgas = 'no')
            tb= np.interp(pb,Model.var[3,:],Model.var[2,:]) # temperature at bottom
            rb = np.interp(pb,Model.var[3,:],Model.var[0,:])
            if( abs(Model.params.cpturb) > 0): # the model includes turbulent pressure
                ptrb = 1. - 1./np.interp(pb,Model.var[3,:],Model.var[20,:])
            else:
                ptrb = 0.
            distb = Aux3DData['distb'] # distance photosphere - bottom
            R3d  =  rb + distb # radius
            logg3d = math.log10(cestam.ggrav) + math.log10(Model.glob[0])  - 2.*math.log10(R3d)
            log_teff3d =( math.log10(Model.glob[2]) - 2.*math.log10(R3d) - math.log10(4.) - math.log10(math.pi) -math.log10(sigma_stefan)) /4.
            teff3d = 10.**log_teff3d

        Outputs = np.zeros(len(Targets) + SeismicConstraints.number)
        log_teff =( math.log10(Model.glob[2]) - 2.*math.log10(Model.glob[1]) - math.log10(4.) - math.log10(math.pi) -math.log10(sigma_stefan)) /4.
        Model.numaxscl = 3104.* ( (Model.glob[0]/cestam.msun) * (cestam.rsun/Model.glob[1])**2) *  math.sqrt(5777./ 10.**log_teff  )  # numax reference value from Mosser, B., Michel, E., Belkacem, K., et al. 2013, A&A, 550, A126

        def Aux3DDataMissing():
            print "Error in OSM/getoutputs: missing auxiliary data"
            
        error = False
        i = 0

        for Target in Targets:
            s = Target.name.lower()
            if(s == 'logg'):
                    Outputs[i] = math.log10(cestam.ggrav) + math.log10(Model.glob[0])  - 2.*math.log10(Model.glob[1])
            elif( s == 'teff'):
                    Outputs[i] = 10.**log_teff 
            elif( s == 'r'):
                    Outputs[i] = Model.glob[1]/cestam.rsun
            elif ( s =='l'):
                    Outputs[i] = Model.glob[2]/cestam.lsun
            elif ( s =='logl'):
                    Outputs[i] = math.log10(Model.glob[2])-math.log10(cestam.lsun)
            elif( s == 'log_teff'):
                    Outputs[i] = log_teff 
            elif( s == 'tb'):
                if(Aux3DData == None):
                    Aux3DDataMissing()
                    error = True
                Outputs[i] = tb
            elif( s == 'logg3d'):
                if(Aux3DData == None):
                    Aux3DDataMissing()
                    error = True
                Outputs[i] = logg3d
            elif( s == 'teff3d'):
                if(Aux3DData == None):
                    Aux3DDataMissing()
                    error = True
                Outputs[i] = teff3d
            elif( s == 'log_teff3d'):
                if(Aux3DData == None):
                    Aux3DDataMissing()
                    error = True
                Outputs[i] = log_teff3d
            elif( s == 'ptrb'):
                if(Aux3DData == None):
                    Aux3DDataMissing()
                    error = True
                Outputs[i] = ptrb
            elif( s == 'deltanuscl'):
                Outputs[i] = 138.8*math.sqrt( (Model.glob[0]/cestam.msun) * (cestam.rsun/Model.glob[1])**3)  # delta_nu reference value (in muHz) from Mosser, B., Michel, E., Belkacem, K., et al. 2013, A&A, 550, A126
            elif( s == 'numaxscl'):
                Outputs[i] = Model.numaxscl
            elif( s == 'largesep'):
                Outputs[i] = largesep(Model,Params,Target,Settings)
            elif( s =='mean_density'):
                Outputs[i] = (Model.glob[0]/cestam.constants.msun)*(cestam.constants.rsun/Model.glob[1])**3
            elif( s =='y_s' ):
                Outputs[i] =  Model.glob[7] # Y at the surface 
            elif( s =='z_s' ):
                Outputs[i] =  1. - Model.glob[6] - Model.glob[7] # Z  at the surface Z = 1 - X - Y
            elif( s =='zsx_s' ):
                Outputs[i] =  (1. - Model.glob[6] - Model.glob[7])/ Model.glob[6]   # Z/X at the surface Z/X = (1 - X - Y)/X
            elif( s =='log_zsx_s' ):
                Outputs[i] =  math.log10((1. - Model.glob[6] - Model.glob[7])/ Model.glob[6])   # log10(Z/X) at the surface Z/X = (1 - X - Y)/X
            else:
                print "Error in OSM/getoutputs: Target named "+  s + " not found"
                error = True
            print ("%s = %g") % (s,Outputs[i])
            i = i + 1

        if(SeismicConstraints.number > 0):
            Seismic_Model = seismic_model(Model,SeismicConstraints,Params,Settings,path='')
            if(Seismic_Model.size != SeismicConstraints.number):
                print Seismic_Model
                print ("a problem occurs with model %s :") % (Model.name)
                print "the theoretical frequencies do not cover the observational range"
                print ("%d seismic constraints are expected while %d were computed") % (SeismicConstraints.number,Seismic_Model.size)
                print "we consider zeros values"
                Outputs[len(Targets):] = 0.
            else:
                Outputs[len(Targets):] = Seismic_Model
        
        return (Outputs,error)
    
    def compute_model(parameters,args,status):
        #   status : status of the model.
        #   status = -1 for the central model to be computed.
        #   status >= 0 for the model computed for the derivative associated with the parameter, which index is  given by the value of status

        error = True
        Models = args[0]
        Targets = args[1]
        index = status + 1
        Model = Models[index]
        name = Model.name
        print '----------------------------------------------------'
        print 'Model # :' , index 
        print 'Name: ' + name
        setupparams(Model,parameters,Settings)
        Model.params.mkdon(replace_eos=False)

        if( (status > -1) and (parameters[status].evol == 1)):
            Model.run.job = 'From previous model'
            Model.run.mod_init = Models[0].name+'_B.rep'
            print '-> evolving model named ' + Models[0].name+'_B.rep'
            (Model)(mkdon=False)
        elif ( (status > -1) and (parameters[status].seismic == 1)):
            # we just copy the central model to compute seismic quantities with different seismic parameters
            Model =  cestam.CModel(Models[0].name)
            Model.name = name
            Models[index] = Model
        else:
            if (model.run.job == 'From PMS'):
                print '-> starting the computation from PMS' 
                Model.run.mod_init = Model.name+'_B.pms'
                (Model)(mkdon=False)
            else:
                print '-> starting the computation from a homogeneous model' 
                Model.run.mod_init = Models[0].name+'_B.hom'
                (Model)(mkdon=False)
            #            print '-> starting the computation from a ZAMS model' 
            #            (Model)(job='zams',mkdon=False,mod_init=Models[0].name+'_B.zams')
        error = (Model.finished == False)
        if(not error):
            # calculation of the eigenfrequencies
            oscprog = 'adipls' # default osc. programme
            if ( Settings['modes'] != None):
                oscprog = Settings['modes']['oscprog']

            if ( oscprog == 'adipls' ):
                if(Aux3DData is not None): # building a patched model
                    print "building a patched model"
                    data,AA,null1,null2,null3,null4 = Aux3DData['ptchfct'](Model.name,Aux3DData['config']) # call to psmlib.patch()
                else:
                    (data,AA) = cesam2amdl(Model)
                error = write_amdl(Model.name+'.amdl',data,AA)
                if(error):
                    print 'Unable to build the amdl file'
                else:
                    os.system('\\rm -f '+ Model.name  + '.gsm')
                    os.system('\\rm -f '+ Model.name  + '.ssm')
                    os.system('\\rm -f '+ Model.name  + '.ef')
                    os.system('\\rm -f '+ Model.name  + '-adipls.log')
                    os.system('runadipls.pl ' + Model.name + '.amdl ' +  Models[0].name+ '.adipls > /dev/null')
            elif (oscprog == 'mad' ):
                if(Aux3DData is not   None):
                    print 'Unable to process MAD code with a patched model'
                cwd = os.getcwd()
                os.system("mad-M " +  Model.name + "-nad.osc  > /dev/null ")
          
        print '----------------------------------------------------'


        print '----------------------------------------------------'
        print 'Model # :' , index 

        if(error == False):
            print '-> Computation successfully finished'
            (Outputs,error) = getoutputs(Model,parameters)
            print '----------------------------------------------------'
            return (Outputs,error)
        else:
            print '-> The model named '+Model.name+' did not converge'
            print '----------------------------------------------------'
            return (np.zeros( len(Targets)) + 1e10,error)

    # --------------------------------------------

    npar = len(Parameters)
    f =open(modelname+'.log','w')

    text = '\n\n'
    text += '===================== OSM started ===============================\n'
   
    model = cestam.CModel(modelname,read=False)
    model.GUI = False
    
    model.finished = False
    model.run.job = 'From ZAMS'
    if( Settings['models'] != None):
        if (  Settings['models']['start'] == 'pms'):
            model.run.c_iben = Settings['models']['cf']
            model.run.job = 'From PMS'
   
    for p in Parameters:
        if( (p.name == 'agemax') or (p.name == 't_stop')):
            p.evol = 1
        if( p.name == 't_stop'):
            model.params.agemax = 1e30
        if( p.name == 'agemax'):
            model.params.t_stop = 1e30
        if ( re.match("^se_",p.name) ): # surface effect parameter -> seismic parameter
            p.seismic = 1

    model.run.type_file='Binary'

    models = [model]
    for i in range(npar):
        s= modelname+("_%3.3i") % (i+1)
        model_im_i = cestam.CModel(s,read=False)
        model_im_i.params.copy(model.params)
        # additional parameters associated with the surface effects
        model_im_i.params.se_a = 0.
        model_im_i.params.se_b = 0.
        model_im_i.params.se_c = 0.
        model_im_i.run.c_iben = model.run.c_iben
        model_im_i.run.job = model.run.job
        model_im_i.params.mkdon(replace_eos=False)
        model_im_i.run.type_file='Binary'
        if( os.access( modelname + '.yl',os.F_OK) ) :
                os.system('\cp -p '+ modelname+ '.yl '  + s + '.yl ')
        #        os.system('cp -f %s' )
        if (model.run.job == 'From PMS'):
            os.system(('cp -f %s_B.pms %s_B.pms') % (modelname,s))
        models.append(model_im_i)

    text += '\nInitial parameters (name = value step min max rate):\n'
    for p in Parameters:
        text +=   ("%s = %g %g %g %g %g\n") % (p.name,p.value,p.step,p.bounds[0],p.bounds[1],p.rate)

    text += 'Targets (name = value sigma):\n'
    for t in Targets:
        text +=  ("%s = %g %g\n") % (t.name,t.value,t.sigma)
    
    nt= len(Targets)
    ny = nt + SeismicConstraints.number
    y = np.empty(ny)
    for i in range(nt):
        y[i] = Targets[i].value
    covar = np.zeros( (ny,ny))
    for i in range(nt):
        covar[i,i] = Targets[i].sigma**2
    if(SeismicConstraints.number > 0):
        y[nt:] =  SeismicConstraints.y[:]
        covar[nt:,nt:] = SeismicConstraints.covar[:,:]

    np.savetxt('covar.txt',covar,fmt='%19.8e')
    
    maxiter = Settings['levmar']['maxiter']
    ftol = Settings['levmar']['ftol'] 
    chi2min = Settings['levmar']['chi2min']
    autostep =  Settings['levmar']['autostep']
    cov_cdtnb_thr =  Settings['levmar']['cov_cdtnb_thr']
    hess_cdtnb_thr = Settings['levmar']['hess_cdtnb_thr']
    
    text += '\n'
    print (text)
    f.write(text)

    sys.stdout.flush()
    text = '\n\n'
  
    (chi2i,parout,iter,outputs,error,msg) =   levmar(modelname,compute_model,Parameters,(models,Targets),y,covar,f,verbose=1,ftol=ftol,maxiter=maxiter,chi2min=chi2min,autostep=autostep,cov_cdtnb_thr=cov_cdtnb_thr,hess_cdtnb_thr=hess_cdtnb_thr)

    if(error):
        raise OSMError("Error in OSM/levmar")

    for i in range(npar):
        s= modelname+("_%3.3i") % (i+1)
        os.system('\\rm -f '+ s + '*')

    text2 = msg
    text2 += '\n'
    text2 +=  ("Chi2 = %g\n") %  (chi2i)
    text2 +=  ("Reduced Chi2 = %g\n\n") %  (chi2i/float(ny))
    if(SeismicConstraints.number > 0):
        chi2s = chi2i
        # we substract the contribution from the non-seismic constraints
        # more straightforward to compute since such constraints are by definition un-correlated
        for i in range(0,nt):
            chi2s -= (outputs[i]-y[i])**2/covar[i,i]
        text2 +=  ("Seismic Chi2 = %g\n") %  (chi2s)
        text2 +=  ("Reduced seismic Chi2 = %g\n\n") %  (chi2s/float(ny-nt))
        text2 +=  ("Non-seismic Chi2 = %g\n") %  (chi2i-chi2s)
        text2 +=  ("Reduced non-seismic Chi2 = %g\n\n") %  ((chi2i-chi2s)/float(nt))
            
    text2 += 'Final parameters (name = value error):\n' 
    for p in parout:
        text2 +=   ("%s = %g %g\n") % (p.name,p.value,p.sigma)

    text2 += '\n'

    text2 += 'Distances to global constraints (name = model data sigma  model-data) :\n'
    for i in range(nt):
        text2 +=  ("%s = %g %g %g %g\n") % ( Targets[i].name,outputs[i],y[i],Targets[i].sigma , outputs[i] - y[i] )
    if(SeismicConstraints.number > 0):
        text2 += "\nDistances to seismic targets (# = model data sigma  model-data)\n"
    for i in range(nt,ny):
        text2 +=  ("%3i = %g %g %g %g\n") % ( i - nt,outputs[i],y[i], math.sqrt(covar[i,i]), outputs[i] - y[i])

    text2 += '\n===================== OSM finished ==============================\n'

    print text2
  
    
    text += text2

    f.write(text)
    f.close()


    return text

        
def read_setup(file):
    try:
        xml = xdm.parse(file)
    except:
        raise OSMError("Error in OSM: some errors occur while reading the XML file, please check its format")

    config =  []

    parameters = []
    targets = []
    
    for s in xml.getElementsByTagName('parameter'):
        parameter = Parameter()
        a = get_first_value('bounds', s).split(',')
        bounds =  [float(a[0]) , float(a[1])]
        parameter.step = float(get_first_value('step', s) )
        parameter.rate = float(get_first_value('rate', s) )
        parameter.bounds = bounds
        parameter.name = s.getAttribute('name')
        parameter.value = float(get_first_value('value', s) )
        parameters.append(parameter)
        if( (parameter.value>  bounds[1]) | (parameter.value <  bounds[0])):
            raise OSMError("Error in OSM: the initial value for %s is outside the bounds" % parameter.name)

    for s in xml.getElementsByTagName('target'):
        target = Target()
        target.name = s.getAttribute('name')
        target.value = float(get_first_value('value', s) )
        target.sigma =  float(get_first_value('sigma', s) )
        targets.append(target)

    xmlsettings = xml.getElementsByTagName('settings')[0]
    
    levmar = xmlsettings.getElementsByTagName('levmar')[0]
    
    autostep = get_first_value('autostep', levmar)
    autostep =  ( int(autostep) == 1) if autostep != None else 0
    cov_cdtnb_thr = get_first_value('cov_cdtnb_thr', levmar)
    cov_cdtnb_thr =  (float(cov_cdtnb_thr)) if  cov_cdtnb_thr != None else 1e13
    hess_cdtnb_thr = get_first_value('hess_cdtnb_thr', levmar)
    hess_cdtnb_thr =  (float(hess_cdtnb_thr)) if  hess_cdtnb_thr != None else 1e13
    settings_levmar = {
        'maxiter': int( get_first_value('maxiter',levmar) ) ,
        'ftol':  float( get_first_value('ftol',levmar) ) ,
        'chi2min': float( get_first_value('chi2min',levmar) ),
        'autostep': autostep,
        'hess_cdtnb_thr' : hess_cdtnb_thr,
        'cov_cdtnb_thr' : cov_cdtnb_thr
        }
    settings_modes = None
    modes = xmlsettings.getElementsByTagName('modes')
    if(len(modes) > 0):
        modes = modes[0]
        l = get_first_value('l', modes)
        nmin = get_first_value('nmin', modes)
        nmax = get_first_value('nmax', modes)
        dn = get_first_value('dn', modes)
        surface_effects =  modes.getElementsByTagName('surface_effects')
        oscprog = get_first_value('oscprog',modes)
        settings_surface_effects = None
        if ( len(surface_effects)  > 0):
            surface_effects = surface_effects[0]
            formula = get_first_value('formula', surface_effects)
            par =  get_first_value('parameters', surface_effects)
            par = np.fromstring(par,sep=',',dtype=np.float) if par != None else None
            numax = get_first_value('numax', surface_effects)
            numax = float(numax)  if numax !=None else None
            settings_surface_effects = { 'formula' : formula , 'parameters': par, 'numax': numax}
            if formula == None:
                raise OSMError("Unspecified formula for the surface effects")
            if (formula.lower() == 'lorentz3' ):
                print 'Modelling the surface effects with a Lorentzian function'
                if (len(par) != 3):
                    raise OSMError("three parameters are expected")
            elif ( (formula.lower() == 'lorentz2')  | (formula.lower() == 'lorentz') ):
                print 'Modelling the surface effects with a Lorentzian function with two parameters'
                if (len(par) != 2):
                    raise OSMError("two parameters are expected")
            elif (formula.lower() == 'kb2008' ):
                print 'Modelling the surface effects using Kjeldsen et Bedding (2008) formula'
                if (len(par) !=2 ):
                    raise OSMError("two parameters are expected")
            elif ( (formula.lower() == 'none') | (formula.lower() == 'no')):
                settings_surface_effects = None
            else:
                raise OSMError("%s: unhandled formula for the surface effects",formula )

        settings_modes = {
            'l': np.fromstring(l,sep=',',dtype=np.int) if l != None else None ,
            'nmin': int(nmin) if nmin != None else None,
            'nmax': int(nmax) if nmax != None else None,
            'dn': int(dn) if dn != None else None,
#            'w2min': float(get_first_value('w2min', modes)) ,
#            'w2max': float(get_first_value('w2max', modes)) ,
            'oscprog': oscprog if oscprog != None else 'adipls',
            'surface_effects': settings_surface_effects
            }

    settings_models = None
    models = xmlsettings.getElementsByTagName('models')
    
    if(len(models) > 0):
        models = models[0]
        start = get_first_value('start', models)
        dy_dz = get_first_value('dy_dz', models)
        yp = get_first_value('yp', models)
        zp = get_first_value('zp', models)
        cf = get_first_value('cf', models) # contraction factor
        start = (clean_string(get_first_value('start', models))).lower()
        if ( (start != 'zams') and (start != 'pms') ):
            raise OSMError("parameter start can only take the value 'zams' or 'pms'")
        settings_models = {
            'dy_dz': float(dy_dz),
            'yp': float(yp), 
            'zp': float(zp),
            'start': start ,
            'cf': float(cf) 
          }
    settings = {
        'levmar': settings_levmar,
        'modes': settings_modes,
        'models': settings_models
        }

    return (config,parameters,targets,settings)


def read_seismic_constraints(file):
    xml = xdm.parse(file)


    xml = xml.getElementsByTagName('seismic_constraints')

    

    if(len(xml)>0):
        xml = xml[0]
        filename =  clean_string(get_first_value('file',xml))
        types =  (re.sub(r'\s', '', clean_string(get_first_value('types',xml)))).split(',')
        matching = 'frequency'
        lfirst = False
        if( len(xml.getElementsByTagName('matching')) > 0 ):
            matching= clean_string(get_first_value('matching',xml))
            print ('matching strategy: ' + matching)
        if( len(xml.getElementsByTagName('lfirst')) > 0 ):
            lfirst = (int(get_first_value('lfirst',xml)) == 1)
            print ('l degree is assumed to be given in the first column')

        seismic_constraints =  SeismicConstraints(filename,types,matching=matching,lfirst=lfirst)
    else:
        seismic_constraints =  SeismicConstraints('')
        
    return seismic_constraints

#-----------------------------------------------------------------
#
#
#-----------------------------------------------------------------
def osminit(name,resume=False):

    print '======================= Initialisation ============================='

    (config,parameters,targets,settings) = read_setup(name+'.xml')

        

    if ( settings['modes'] != None):
        oscprog = settings['modes']['oscprog']
        if ( (oscprog != 'adipls') & (oscprog != 'mad') ):
            raise OSMError( ("%s : unknown oscillations programme" % oscprog) )

    seismic_constraints = read_seismic_constraints(name+'.xml')

    if (  not(os.access(name+'.don',os.F_OK)) ):
        raise OSMError("Error in OSM: missing file "+ name+".don")

    refmodel =  cestam.CModel(name)
    setupparams(refmodel,parameters,settings)

    OutputPath = name+'/'

    if (  not(os.access(OutputPath,os.F_OK))):
        os.mkdir(OutputPath)

    if( os.access(name + '.mix',os.F_OK) ) :
        os.system('\cp -p '+ name+ '.mix ' + OutputPath + '/mixture')
    elif( os.access( 'mixture',os.F_OK) ) :
        os.system('\cp -p mixture ' + OutputPath)

    if( os.access(name + '.rg',os.F_OK) ) :
        os.system('\cp -p '+ name+ '.rg ' + OutputPath + '/reglages')
    elif( os.access( 'reglages',os.F_OK) ) :
        os.system('\cp -p reglages ' + OutputPath)

    if( os.access(name + '.yl',os.F_OK) ) :
        os.system('\cp -p '+ name+ '.yl ' + OutputPath + '/'  + name+ '.yl ')

    adiplsin = name + '.adipls'

    if( os.access(adiplsin,os.F_OK) ) :
        os.system('\cp -p '+  adiplsin + '  ' +  OutputPath + '.')
    else:
        os.system('\cp -p  adipls.in ' + OutputPath +  adiplsin)

    os.system('\cp -p '+ name+ '.xml ' + OutputPath + '/.')

    os.chdir(OutputPath)
    
    if(resume):
        print ('Initial parameters taken from previous calculation')
        paramprev = np.loadtxt('param.txt',usecols=(1,))
        i = 0
        for p in parameters:
            p.value = paramprev[i]
            i += 1    

    model = cestam.CModel(name,read=False)

    model.params.copy(refmodel.params)
    
    if( refmodel.params.diffusion ):
        print 'Diffusion included'
##         model.params.x0 = (1. - model.params.y0)/(1. + model.params.zsx0)
##         model.params.z0 = (1. - model.params.x0 - model.params.y0)
    
##    setupparams(model,parameters,settings)
##    print 'x0,y0,z0,zsx0 = ',model.params.x0, model.params.y0,model.params.z0,model.params.zsx0
    model.params.agemax = 0.
    model.params.t_stop = 1E30
    
    noguess = False
    for p in parameters:
        if(p.name == 'agemax' or p.name == 't_stop'):
            if(p.value < 0):
                noguess = True
        if(noguess):
            # we chech that Teff is among the targets
            teff = -1.
            for t in targets:
                if(t.name == 'teff' or t.name == 'log_teff'):
                    if(t.name == 'teff'):
                        teff = t.value
                    else:
                        teff = 10.**(t.value)
            if(teff < 0.):
                raise OSMError("Teff must be among the targets when there is no guess value for agemax or x_stop or t_stop")
                    

    model.params.mkdon(replace_eos=False)
    model.GUI = False

    start = 'zams'
    cf = 8e-5
    model.finished = False
    if( settings['models'] != None):
        start = settings['models']['start']
        cf = settings['models']['cf']
    if (start == 'zams'):
        if( (os.access(name+'_B.hom',os.F_OK) ) ):
            print '\na homogeneous model is already present in the W.D., we try to start from it'
            model.params.agemax = 0.
            model.run.type_file= 'Binary'
            model.run.job = 'From ZAMS'
            model.run.mod_init = './'+ name+'_B.hom'
            model(mkdon=False)

        if(model.finished == False) :
            zamsmodel(model,model.params.mtot)
        model.run.type_file= 'Binary'
        model.params.mkdon(replace_eos=False)
    if (start == 'pms'):
        model.run.c_iben = cf
        if( (os.access(name+'_B.pms',os.F_OK) ) ):
            print '\nan initial PMS model is already present in the W.D., we try to start from it'
            model.params.agemax = 0.
            model.run.type_file= 'Binary'
            model.run.job = 'From PMS'            
            model.run.mod_init = './'+ name+'_B.pms'         
            model(mkdon=False)
        if(model.finished == False) :
            pmsmodel(model,model.params.mtot)

    if(model.finished == False):
        raise OSMError("Error in OSM: unable to compute the initial model")
        
    if (start == 'zams'):
        os.system('\cp -f ' + name+ '_B.dat' + ' ' + name + '_B.zams')

    if(noguess):
        print '\nno guess value is given for age or T_STOP'
        OK  = False
        teff = 1e10
        for T in targets:
            if T.name.lower() == 'teff':
                teff = T.value
                OK = True
            if T.name.lower() == 'log_teff':
                teff = 10.**T.value
                OK = True
        if(not OK):
            raise   OSMError("Error in OSM: when no guess value is provided, Teff must be among the targets")
        model.params.t_stop = 1E30
        model.params.agemax = 15E3
        model.run.type_file= 'Binary'
        print ('\ncomputing a model that matches approximately Teff= %f K' % teff)
        for p in parameters:
            if(p.name == 'agemax'):
                print ("we start with agemax =  %f") % (p.bounds[0])
                model.params.agemax = p.bounds[0]    
            if(p.name == 't_stop'):
                model.params.t_stop = p.bounds[0]    
                print ("we start with T_STOP = %e") % (p.bounds[0])
        model.params.log_teff = 10
        model.params.mkdon(replace_eos=False)
        if (start == 'zams'):
            model.run.job = 'From ZAMS'
            model.run.mod_init = name+'_B.hom'
            model(mkdon=False)
        else:
            model.run.job = 'From PMS'
            model.run.mod_init = './'+ name+'_B.pms'
            model(mkdon=False)
        model.params.t_stop = 1E30
        model.params.agemax = 15E3
        model.params.log_teff = - math.log10(teff)
        model.params.mkdon(replace_eos=False)
        model.run.job = 'From previous model'
        model.run.mod_init = name+'_B.rep'
        model(mkdon=False)
        model.params.log_teff = 10
        model.params.mkdon(replace_eos=False)
        if(model.finished == False):
            raise OSMError("Error in OSM: unable to compute the guess model")
        for p in parameters:
            if(p.name == 'agemax'):    
                p.value = model.glob[10]    
            if(p.name == 't_stop'):
                model.read_osc()    
                p.value = model.var[2,model.ntot-1] # temperature at the center

    os.chdir("../.")

    print '======================= Initialisation FINISHED ===================='

    return parameters,targets,settings,seismic_constraints

def osmrun(name,parameters,targets,settings,seismic_constraints,resume=False):


    OutputPath = name+'/'
    
    os.chdir(OutputPath)

    results = compute_optimal(name,parameters,targets,settings,seismic_constraints,resume=False)

    os.chdir("../.")

    return results



"""
Optimal Stellar Models (OSM)

Copyright (c) 2012 R. Samadi (LESIA - Observatoire de Paris)

This is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
 
This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this code.  If not, see <http://www.gnu.org/licenses/>.
"""

from distutils.core import setup



setup(name = 'osm',
      version = '1.4',
      description = 'Optimal Stellar Models (OSM): a Python module for calculating optimal stellar models with CESTAM',
      author = 'R. Samadi',
      author_email = 'reza.samadi@obspm.fr',
      url =  'http://lesia.obspm.fr/',
      scripts = ['osm.py','runadipls.pl'],
      py_modules = ['osmlib','osmmodes','osmlevmar'],
      long_description_content_type= 'text/markdown',
      long_description = open('README.txt').read()
      )



"""
Optimal Stellar Models (OSM)

Copyright (c) 2012 R. Samadi (LESIA - Observatoire de Paris)

This is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
 
This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this code.  If not, see <http://www.gnu.org/licenses/>.
"""

import matplotlib.pyplot as plt
import numpy as np
import cestam.FortranIO
import math
from osmlib import *

msun = 1.98919e33           # solar mass
rsun = 6.9599e10            # solar radius
gmsun = 1.32712438e26       # G Msun
ggrav = gmsun/msun          # grav. constant G

class OSMError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value) 

## def dnu(modes):
##     # INPUTS:
##     # modes[:,0] : n order
##     # modes[:,1] : l degree
##     # modes[:,2] : nu frequency muHz
##     # modes[:,3] : sigma error in frequency mHz
##     # OUTPUTS:
##     # (y,covar):
##     #  y : Delta_nu [in muHz] as a function of frequency
##     #

##     l = modes[:,1]
##     nu = modes[:,2]

##     Nf =   len(nu)
##     # number of modes for each l degree
##     Nl = []
##     m = 1
##     for i in range(1,Nf):
##         if( l[i] != l[i-1]):
##             Nl.append(m)
##             m = 0
##         m += 1
##     Nl.append(m)
##     N = Nf - len(Nl)
##     coef = np.zeros( (N,Nf) ) #
##     m = 0
##     n = 0
##     for k in range(len(Nl)): # loop over the l values:
##         for p in range(Nl[k]-1):
##             for i in range(Nl[k]):
##                 if( i == p):
##                     coef[m+p, n + i] = -1.
##                 elif( i  == p + 1 ):
##                     coef[m+p, n + i] = 1.
##         m += Nl[k]-1
##         n += Nl[k]
##     y = np.empty(N)
##     for i in range(N):
##         y[i] = (coef[i,:]*nu).sum()
##     return (y,coef)

def dnu(modes,lval=None):
    # Delta_nu (n) = nu_{n,l} - nu_{n-1,l} 
    # 
    # INPUTS:
    # modes[:,0] : n order
    # modes[:,1] : l degree
    # modes[:,2] : nu frequency muHz
    # modes[:,3] : sigma error in frequency mHz
    # OUTPUTS:
    # (y,coef):
    #  y : Delta_nu [in muHz] as a function of frequency
    #

    if(lval == None):
        lval = get_l_values(modes)

    l = modes[:,1]
    nu = modes[:,2]
    n =  modes[:,0]
    
    Nf =   len(nu)
    
    # number of modes for each l degree
    Nl = []
    N = 0
    for s in lval:
        j = np.where(l == s)
        m = len(j[0])
        Nl.append(m)
        N += m-1

    coef = np.zeros((N,Nf))  
    yn = np.empty(N)
    p = 0 
    for k in range(len(lval)):
        if(Nl[k] > 1):
            i = np.where( l ==  lval[k] ) [0]
            for m in i[1:]:
                for j in range(Nf):
                    if( n[j] == n[m] and l[j] == lval[k] ):
                        coef[p,j] = 1.
                        yn[p] = n[j]
                    if( n[j] == n[m]-1 and l[j] == lval[k]):
                        coef[p,j] = -1.
                p +=  1

    y = np.empty(N)
    for i in range(N):
        y[i] = (coef[i,:]*nu).sum()
    return (y,coef,yn)

def d01(modes):
    # d01 = nu_{n,0} - ( nu_{n-1,1} + nu_{n,1}) / 2
    # INPUTS:
    # modes[:,0] : n order
    # modes[:,1] : l degree
    # modes[:,2] : nu frequency muHz
    # modes[:,3] : sigma error in frequency mHz
    # OUTPUTS:
    # (y,coef):
    #  y : d01 [in muHz] as a function of frequency
    #

    n = modes[:,0]
    l = modes[:,1]
    nu = modes[:,2]

    Nf =   len(nu)
 
    i0 = np.where(l == 0)
    i1 = np.where(l == 1)

    i0 = i0[0]
    i1 = i1[0]
    


    if( l[i0[0]] !=0 and l[i1[0]] != 1):
        print 'error in d01, the l=1 modes must be placed after the l=0 modes'
        return

    n0min = np.min(n[i0])
    n0max = np.max(n[i0])
    n1min = np.min(n[i1])
    n1max = np.max(n[i1])

    n0 = int(max([n0min,n1min+1]))
    n1 = int(min([n0max,n1max]))

    i00 = i0[np.argmin(np.abs(n[i0]-n0))]
    i01 = i0[np.argmin(np.abs(n[i0]-n1))]
   
    #    print n0,n1,i00,i01
    N = n1 - n0 +1
    yn = np.empty(N)
    coef = np.zeros((N,Nf))
    for i in range(i00,i01+1):
        #   print n[i],l[i]
        for j in range(Nf):
            if( n[j] == n[i] and  l[j] == 0 ):
                coef[i,j] = 1.
                yn[i] = n[j]
            if( (n[j] == n[i]-1 or n[j] == n[i])  and  l[j] == 1 ):
                coef[i,j] = -1./2.
        
    y = np.empty(N)
    for i in range(N):
        y[i] = (coef[i,:]*nu).sum()

    return (y,coef,yn)



def d02(modes):
    # d02 = nu_{n,0} - nu_{n-1,2} 
    # INPUTS:
    # modes[:,0] : n order
    # modes[:,1] : l degree
    # modes[:,2] : nu frequency muHz
    # modes[:,3] : sigma error in frequency mHz
    # OUTPUTS:
    # (y,coef):
    #  y : d02 [in muHz] as a function of frequency
    #

    n = modes[:,0]
    l = modes[:,1]
    nu = modes[:,2]

    Nf =   len(nu)
 
    i0 = np.where(l == 0)[0]
    i1 = np.where(l == 2)[0]

    if( l[i0[0]] !=0 and l[i1[0]] != 2):
        print 'error in d02, the l=2 modes must be placed after the l=0 modes'
        return

    n0min = np.min(n[i0])
    n0max = np.max(n[i0])
    n1min = np.min(n[i1])
    n1max = np.max(n[i1])

    n0 = int(max([n0min,n1min+1]))
    n1 = int(min([n0max,n1max+1]))

    
    N = n1 - n0 +1
    yn = np.empty(N)
    coef = np.zeros((N,Nf))
    for i in range(N):
        #   print n[i],l[i]
        for j in range(Nf):
            if( n[j] == i + n0 and  l[j] == 0 ):
                coef[i,j] = 1.
                yn[i] = n[j]
            if( (n[j] == i + n0 - 1)  and  l[j] == 2 ):
                coef[i,j] = -1.
        
    y = np.empty(N)
    for i in range(N):
        y[i] = (coef[i,:]*nu).sum()

    return (y,coef,yn)
        
def sd(modes,lval=None):
    # second difference as a function of frequency (Gough 1990, Houdek & Gough 2007)
    # sd = nu_{n-1,l} - 2 nu_{n,l} + nu_{n+1,l}
    # 
    # INPUTS:
    # modes[:,0] : n order
    # modes[:,1] : l degree
    # modes[:,2] : nu frequency muHz
    # modes[:,3] : sigma error in frequency mHz
    # OUTPUTS:
    # (y,coef):
    #  y : second difference [in muHz] as a function of frequency

    if(lval == None):
        lval = get_l_values(modes)

    l = modes[:,1]
    nu = modes[:,2]
    n =  modes[:,0]
    
    Nf =   len(nu)
    
    # number of modes for each l degree
    Nl = []
    N = 0
    for s in lval:
        j = np.where(l == s)
        m = len(j[0])
        Nl.append(m)
        N += m-2

    coef = np.zeros((N,Nf))  
    yn = np.empty(N)
    p = 0 
    for k in range(len(lval)):
        if(Nl[k] > 1):
            i = np.where( l ==  lval[k] ) [0]
            for m in i[1:-1]:
                for j in range(Nf):
                    if( n[j] == n[m] and l[j] == lval[k] ):
                        coef[p,j] = -2.
                        yn[p] = n[j]
                    if( n[j] == n[m]-1 and l[j] == lval[k]):
                        coef[p,j] = 1.
                    if( n[j] == n[m]+1 and l[j] == lval[k]):
                        coef[p,j] = 1.
                p +=  1

    y = np.empty(N)
    for i in range(N):
        y[i] = (coef[i,:]*nu).sum()
    return (y,coef,yn)

   
def sd01(modes):
    # sd01 second difference 01 as a function of frequency as defined in Eq. (4) of Roxburgh & Vorontsov (2003,A&A)
    #
    # sd01 (n) = (1/8) ( nu_{n-1,0} - 4 nu_{n-1,1} + 6 nu_{n,0} - 4 nu_{n,1} + nu_{n+1,0})
    # 
    # INPUTS:
    # modes[:,0] : n order
    # modes[:,1] : l degree
    # modes[:,2] : nu frequency muHz
    # modes[:,3] : sigma error in frequency mHz
    # OUTPUTS:
    # (y,coef):
    #  y : sd01 second difference [in muHz] 
    #

    n = modes[:,0]
    l = modes[:,1]
    nu = modes[:,2]

    Nf =   len(nu)
 
    i0 = np.where(l == 0)[0]
    i1 = np.where(l == 1)[0]

    if( l[i0[0]] !=0 and l[i1[0]] != 2):
        print 'error in sd01, the l=1 modes must be placed after the l=0 modes'
        return

    n0min = np.min(n[i0])
    n0max = np.max(n[i0])
    n1min = np.min(n[i1])
    n1max = np.max(n[i1])

    n0 = int(max([n0min,n1min])) + 1
    n1 = int(min([n0max,n1max])) - 1
    #    print n0,n1,i00,i01

    N = n1 - n0 +1
    if( N < 1):
        print 'error in sd01, not enough l=0 and/or l=1 modes'
        return
    yn = np.empty(N)    
    coef = np.zeros((N,Nf))
    for i in range(N):
        #   print n[i],l[i]
        for j in range(Nf):
            if( n[j] == i + n0 -1  and  l[j] == 0 ):
                coef[i,j] = 1./8
            if( n[j] == i + n0   and  l[j] == 0 ):
                coef[i,j] = 6./8
                yn[i] = n[j]
            if( n[j] == i + n0+1   and  l[j] == 0 ):
                coef[i,j] = 1./8
            if( n[j] == i + n0 -1   and  l[j] == 1 ):
                coef[i,j] = -0.5
            if( n[j] == i + n0    and  l[j] == 1 ):
                coef[i,j] = -0.5
        
    y = np.empty(N)
    for i in range(N):
        y[i] = (coef[i,:]*nu).sum()

    return (y,coef,yn)

def sd10(modes):
    # sd10 second difference "10" as a function of frequency as defined in Eq. (5) of Roxburgh & Vorontsov (2003,A&A)
    #
    # sd10 (n) = -(1/8) ( nu_{n-1,1} - 4 nu_{n,0} + 6 nu_{n,1} - 4 nu_{n+1,0} + nu_{n+1,1})
    #
    #
    # INPUTS:
    # modes[:,0] : n order
    # modes[:,1] : l degree
    # modes[:,2] : nu frequency muHz
    # modes[:,3] : sigma error in frequency mHz
    # OUTPUTS:
    # (y,coef):
    #  y : sd10 second difference [in muHz] 
    #

    n = modes[:,0]
    l = modes[:,1]
    nu = modes[:,2]

    Nf =   len(nu)
 
    i0 = np.where(l == 0)[0]
    i1 = np.where(l == 1)[0]

    if( l[i0[0]] !=0 and l[i1[0]] != 2):
        print 'error in sd10, the l=1 modes must be placed after the l=0 modes'
        return

    n0min = np.min(n[i0])
    n0max = np.max(n[i0])
    n1min = np.min(n[i1])
    n1max = np.max(n[i1])

    n0 = int(max([n0min,n1min])) + 1
    n1 = int(min([n0max,n1max])) - 1
    #    print n0,n1,i00,i10

    N = n1 - n0 +1
    if( N < 1):
        print 'error in sd10, not enough l=0 and/or l=1 modes'
        return
    yn = np.empty(N)    
    coef = np.zeros((N,Nf))
    for i in range(N):
        #   print n[i],l[i]
        for j in range(Nf):
            if( n[j] == i + n0 -1  and  l[j] == 1 ):
                coef[i,j] = -1./8
            if( n[j] == i + n0   and  l[j] == 1 ):
                coef[i,j] = -6./8
                yn[i] = n[j] 
            if( n[j] == i + n0+1   and  l[j] == 1 ):
                coef[i,j] = -1./8
            if( n[j] == i + n0    and  l[j] == 0 ):
                coef[i,j] = 0.5
            if( n[j] == i + n0 +1    and  l[j] == 0 ):
                coef[i,j] = 0.5
        
    y = np.empty(N)
    for i in range(N):
        y[i] = (coef[i,:]*nu).sum()

    return (y,coef,yn)

def rsd01(modes):
    # ratio sd01(n) / dnu1(n)
    # as defined in Eq. 1 of Lebreton & Goupil (2012, A&A)
    # with dnu1 = nu_{n,1} - nu_{n-1,1}
    n = modes[:,0]
    l = modes[:,1]
    nu = modes[:,2]

    Nf =   len(nu)
 
    i0 = np.where(l == 0)[0]
    i1 = np.where(l == 1)[0]

    if( l[i0[0]] !=0 and l[i1[0]] != 2):
        print 'error in rsd01, the l=1 modes must be placed after the l=0 modes'
        return

    n0min = np.min(n[i0])
    n0max = np.max(n[i0])
    n1min = np.min(n[i1])
    n1max = np.max(n[i1])

    n0 = int(max([n0min,n1min])) + 1
    n1 = int(min([n0max,n1max])) - 1
    #    print n0,n1,i00,i01

    N = n1 - n0 +1
    if( N < 1):
        print 'error in rsd01, not enough l=0 and/or l=1 modes'
        return
    yn = np.empty(N) 
    coef = np.zeros((N,Nf))
    coefd = np.zeros((N,Nf))
    for i in range(N):
        #   print n[i],l[i]
        for j in range(Nf):
            # sd01
            if( n[j] == i + n0 -1  and  l[j] == 0 ):
                coef[i,j] = 1./8
            if( n[j] == i + n0   and  l[j] == 0 ):
                coef[i,j] = 6./8
                yn[i] = n[j]
            if( n[j] == i + n0+1   and  l[j] == 0 ):
                coef[i,j] = 1./8
            if( n[j] == i + n0 -1   and  l[j] == 1 ):
                coef[i,j] = -0.5
            if( n[j] == i + n0    and  l[j] == 1 ):
                coef[i,j] = -0.5
            # dnu1
            if( n[j] == i + n0 -1   and  l[j] == 1 ):
                coefd[i,j] = -1.
            if( n[j] == i + n0    and  l[j] == 1 ):
                coefd[i,j] = 1.
        
    y = np.empty(N)
    sd01 = np.empty(N)
    dnu1 =np.empty(N)
    coefy = np.zeros((N,Nf)) 
    for i in range(N):
        sd01[i] = (coef[i,:]*nu).sum()
        dnu1[i] = (coefd[i,:]*nu).sum()
        y[i] = sd01[i]/dnu1[i]
        coefy[i,:] = y[i]* ( coef[i,:]/sd01[i]  - coefd[i,:]/dnu1[i])

    return (y,coefy,coef,coefd,yn)

def rsd10(modes):
    # ratio sd10 (n) / dnu0 (n)
    # as defined in Eq. 1 of Lebreton & Goupil (2012, A&A)
    # with dnu0 =  nu_{n+1,0} - nu_{n,0}
    n = modes[:,0]
    l = modes[:,1]
    nu = modes[:,2]

    Nf =   len(nu)
 
    i0 = np.where(l == 0)[0]
    i1 = np.where(l == 1)[0]

    if( l[i0[0]] !=0 and l[i1[0]] != 2):
        print 'error in rsd10, the l=1 modes must be placed after the l=0 modes'
        return

    n0min = np.min(n[i0])
    n0max = np.max(n[i0])
    n1min = np.min(n[i1])
    n1max = np.max(n[i1])

    n0 = int(max([n0min,n1min])) + 1
    n1 = int(min([n0max,n1max])) - 1
    #    print n0,n1,i00,i01

    N = n1 - n0 +1
    if( N < 1):
        print 'error in rsd10, not enough l=0 and/or l=1 modes'
        return
    yn = np.empty(N) 
    coef = np.zeros((N,Nf))
    coefd = np.zeros((N,Nf))
    for i in range(N):
        #   print n[i],l[i]
        for j in range(Nf):
            # sd10
            if( n[j] == i + n0 -1  and  l[j] == 1 ):
                coef[i,j] = -1./8
            if( n[j] == i + n0   and  l[j] == 1 ):
                coef[i,j] = -6./8
                yn[i] = n[j]
            if( n[j] == i + n0+1   and  l[j] == 1 ):
                coef[i,j] = -1./8
            if( n[j] == i + n0    and  l[j] == 0 ):
                coef[i,j] = 0.5
            if( n[j] == i + n0 +1    and  l[j] == 0 ):
                coef[i,j] = 0.5
            # dnu0
            if( n[j] == i + n0    and  l[j] == 0 ):
                coefd[i,j] = -1.
            if( n[j] == i + n0+1    and  l[j] == 0 ):
                coefd[i,j] = 1.
        
    y = np.empty(N)
    sd10 = np.empty(N)
    dnu0 =np.empty(N)
    coefy = np.zeros((N,Nf)) 
    for i in range(N):
        sd10[i] = (coef[i,:]*nu).sum()
        dnu0[i] = (coefd[i,:]*nu).sum()
        y[i] = sd10[i]/dnu0[i]
        coefy[i,:] = y[i]* ( coef[i,:]/sd10[i]  - coefd[i,:]/dnu0[i])

    return (y,coefy,coef,coefd,yn)
    
def rd02(modes):
    # ratio d02 (n) / dnu1 (n)
    # with dnu1 =  nu_{n,1} - nu_{n-1,1}
    n = modes[:,0]
    l = modes[:,1]
    nu = modes[:,2]

    Nf =   len(nu)
 
    i0 = np.where(l == 0)[0]
    i1 = np.where(l == 1)[0]
    i2 = np.where(l == 2)[0]

    if( l[i0[0]] !=0 and l[i1[0]] != 1 and l[i2[0]] != 2):
        print 'error in rd02, the l=0, l=1 and l=2 modes must placed in increasing l degree'
        return

    n0min = np.min(n[i0])
    n0max = np.max(n[i0])
    n1min = np.min(n[i1])
    n1max = np.max(n[i1])
    n2min = np.min(n[i2])
    n2max = np.max(n[i2])

    n0 = int(max([n0min,n1min+1,n2min+1])) 
    n1 = int(min([n0max,n1max+1,n2max+1])) 
    #    print n0,n1,i00,i01

    N = n1 - n0 +1
    if( N < 1):
        print 'error in rd02, not enough l=0, l=1,  and l=2 modes'
        return
    yn = np.empty(N) 
    coef = np.zeros((N,Nf))
    coefd = np.zeros((N,Nf))
    for i in range(N):
        #   print n[i],l[i]
        for j in range(Nf):
            # d02
            if( n[j] == i + n0 and  l[j] == 0 ):
                coef[i,j] = 1.
            if( (n[j] == i + n0 - 1)  and  l[j] == 2 ):
                coef[i,j] = -1.
            # dnu1
            if( n[j] == i + n0 -1    and  l[j] == 1 ):
                coefd[i,j] = -1.
            if( n[j] == i + n0    and  l[j] == 1 ):
                coefd[i,j] = 1.
                yn[i] = n[j]
        
    y = np.empty(N)
    d02 = np.empty(N)
    dnu1 =np.empty(N)
    coefy = np.zeros((N,Nf)) 
    for i in range(N):
        d02[i] = (coef[i,:]*nu).sum()
        dnu1[i] = (coefd[i,:]*nu).sum()
        y[i] = d02[i]/dnu1[i]
        coefy[i,:] = y[i]* ( coef[i,:]/d02[i]  - coefd[i,:]/dnu1[i])

    return (y,coefy,coef,coefd,yn)



class SeismicConstraints:

    file = ''
    type = ''
    matching = 'frequency'
    n = np.empty(0,dtype=np.int)
    l = np.empty(0,dtype=np.int)
    nu = np.empty(0)
    sigma = np.empty(0)
    y = np.empty(0)
    yn = np.empty(0)
    covar = np.empty(0)
    coef = np.empty(0)
    number = 0
    
    def __init__(self,file='',types=['nu'],matching = 'frequency',data=None,lfirst=False):
        self.file = file
        self.types = types
        if (file != ''):
            print 'we read the table of frequencies: '+ file
            data = np.loadtxt(file,comments='#')        
        else:
            if (data == None):
                return
        if(lfirst):
            self.n = np.array(data[:,1],dtype=np.int)
            self.l = np.array(data[:,0],dtype=np.int)
        else:
            self.n = np.array(data[:,0],dtype=np.int)
            self.l = np.array(data[:,1],dtype=np.int)
        self.nu =  data[:,2]
        self.sigma = data[:,3]
        self.lval = np.unique(self.l)
        #            raise OSMError("Error in OSM / SeismicConstraints: unspecified file")
        if ( (matching != 'frequency') &  ( matching != 'order') & (matching != 'continuous_frequency') &  ( matching != 'continuous_order') ):
            raise OSMError("Error in OSM / SeismicConstraints: unhandled option matching : " +  matching)
        self.matching = matching
        print ('total number of frequencies: %d') % (len(self.nu))
        print 'n   l  nu  sigma'
        for i in range(len(self.nu)):
            print ("%d %d %f %f") % (self.n[i],self.l[i],self.nu[i],self.sigma[i])
        if ( self.nu.size < 1 ):
            raise OSMError("Error in OSM / SeismicConstraints: more than one frequency is required" )
        self.coef = np.empty( (0,self.nu.size))
        for type in types:
            if( type == 'dnu'):
                (y,coef,yn) = dnu(data)
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'dnu0'):
                (y,coef,yn) = dnu(data,lval=[0])
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'dnu1'):
                (y,coef,yn) = dnu(data,lval=[1])
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'dnu2'):
                (y,coef,yn) = dnu(data,lval=[2])
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'd01'):
                (y,coef,yn) = d01(data)
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'd02'):
                (y,coef,yn) = d02(data)
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'rd02'):
                (y,coef,a,b,yn) = rd02(data)
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'sd'):
                (y,coef,yn) = sd(data)
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'sd01'):
                (y,coef,yn) = sd01(data)
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'sd10'):
                (y,coef,yn) = sd10(data)
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'rsd10'):
                (y,coef,a,b,yn) = rsd10(data)
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'rsd01'):
                (y,coef,a,b,yn) = rsd01(data)
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,yn)
                self.coef = np.append(self.coef,coef,axis=0)
            elif( type == 'nu'):
                y = np.empty(self.nu.size)
                coef = np.zeros((self.nu.size,self.nu.size))
                for i in range(self.nu.size):
                    y[i] = self.nu[i]
                    coef[i,i] = 1
                self.y = np.append(self.y,y)
                self.yn = np.append(self.yn,self.n)
                self.coef = np.append(self.coef,coef,axis=0)
            else:
                raise OSMError("Error in OSM / SeismicConstraints:  seimic consraint of type '"+ type + "' is not handled")
        nc = self.coef.shape[0]
        self.covar = np.empty( (nc,nc) )
        s = ''
        for i in range(nc):
            for j in range(nc):
                self.covar[i,j] = (self.coef[i,:] * self.coef[j,:] * self.sigma **2 ).sum()
                s += (("%g ") % (self.covar[i,j]))
            s+= '\n'
        #plt.imshow(self.covar)
        #plt.show(block=True)
        self.number = nc
        print 'type(s) of seismic constraints:'
        for t in types:
            print t
        print 'total number of seismic constraints:', self.number
        print 'seismic constraints (# =  value sigma):'
        for i in range(self.number):
            print ("%3i = %f %f") % (i,self.y[i],math.sqrt(self.covar[i,i]))
        print ("Covariance matrix:")
        print (s)


def read_mad(file):
    '''
     OUTPUTS:
     modes[:,0] : n order
     modes[:,1] : l degree
     modes[:,2] : nu frequency in muHz (eigenvalue)
     modes[:,3] : square normalised frequency (w2)
     modes[:,4] : Richardson frequency in muHz
     '''
    data = np.loadtxt(file)

    modes = np.zeros( (data.shape[0],4) )

    modes[:,0] = data[:,1] #  n order
    modes[:,1] = data[:,0] #  l degree
    modes[:,2] = data[:,11] # nu frequency in muHz
    modes[:,3] = data[:,5]**2 # square normalised frequency (w2)

    return modes
    
def read_agsm(file,G=ggrav):
    '''	
     OUTPUTS:
     modes[:,0] : n order
     modes[:,1] : l degree
     modes[:,2] : nu frequency in muHz (eigenvalue)
     modes[:,3] : square normalised frequency (w2)
     modes[:,4] : Richardson frequency in muHz
     modes[:,5] : variationnal frequency in muHz
     modes[:,6] : icase
    '''

    f = cestam.FortranIO.FortranBinaryFile(file, mode='r')

    cs = []
    while True:
        try:
            csi = f.readRecordNative('d')
            cs.append(csi)
        except IndexError:
            break

    nmodes = len(cs)
    modes = np.zeros((nmodes,7))
    mstar = (cs[0])[1]
    rstar = (cs[0])[2]
    for i in range(nmodes):
        modes[i,0] = (cs[i]) [18] # n
        modes[i,1] = (cs[i]) [17] # l        
        sigma2 = (cs[i]) [19]
        omega2 = sigma2*G*mstar/(rstar**3)
        modes[i,2] = math.sqrt(omega2)/(2.*math.pi)*1e6 # nu in muHz
        modes[i,3] = sigma2
        modes[i,4] = (cs[i]) [36]*1e3 # Richardson frequency in muHz
        modes[i,5] = (cs[i]) [26]*1e3 # variational frequency in muHz

    f.close()
    # reading the 'ics' integer variables
    f = cestam.FortranIO.FortranBinaryFile(file, mode='r')
    n = 38*2
    for i in range(nmodes):
        data = f.readRecordNative('i')
        ics = data[n:n+8]
        modes[i,6] = ics[4]

    f.close()

    return modes

def read_amde(name):
    '''
     read eigenfunctions from file.
     assumes that the file contains a full set of eigenfunctions
     return (x,y,modes) where:
     x[n] : radial mesh (r/R)
     y[n,0:6,k] : the eigenfunctions
     modes[k,0:7] : modes properties (see read_agsm)
     k is the modes index
     n is the mesh index
    '''
    
    f = open(name, 'r')
    cs = []
    ics = []
    while True:
        try:
            size = np.fromstring(f.read(4),'i') [0]  # record size in bytes
            record = f.read(size)
            f.read(4)
            n = 38*8
            csi = np.fromstring(record[0:n],'d')
            icsi = np.fromstring(record[n:n+8*4],'i')
            n = 50*8
            nn =    np.fromstring(record[n:n+4],'i')[0]
            cs.append(csi)
            ics.append(icsi)
        except IndexError:
            break
    f.close()
    nmodes = len(cs)
    modes = np.zeros((nmodes,7))
    mstar = (cs[0])[1]
    rstar = (cs[0])[2]
    for i in range(nmodes):
        modes[i,0] = (cs[i]) [18] # n
        modes[i,1] = (cs[i]) [17] # l        
        sigma2 = (cs[i]) [19]
        omega2 = sigma2*ggrav*mstar/(rstar**3)
        modes[i,2] = math.sqrt(omega2)/(2.*math.pi)*1e6 # nu in muHz
        modes[i,3] = sigma2
        modes[i,4] = (cs[i]) [36]*1e3 # Richardson frequency in muHz
        modes[i,5] = (cs[i]) [26]*1e3 # variational frequency in muHz
        modes[i,6] = (ics[i])[4]
    f = open(name, 'r')
    y = np.empty( (nn,6,nmodes) )
    x = np.empty( (nn) )
    for i in range(nmodes):
            size = np.fromstring(f.read(4),'i') [0]
            record = f.read(size)
            f.read(4)
            n = 50*8 + 4
            for k in range(nn):
                yi = np.fromstring(record[n:n+7*8],'d')
                n += 7*8
                y[k,:,i] = yi[1:]
                x[k] = yi[0]
    f.close()
    return (x,y,modes)

def surface_effects(nu,Model,Params,Settings):
    if( Settings['modes'] is not None ):
        surface_effects = Settings['modes']['surface_effects']
        if(surface_effects is not None ): # adding surface effects
            par = surface_effects['parameters'].copy()
            for s in Params:
                if (s.name == 'se_a'):
                    par[0] = s.value
                if (s.name == 'se_b'):
                    par[1] = s.value
                if (s.name == 'se_c'):
                    par[2] = s.value
            formula = surface_effects['formula']
            if( surface_effects['numax'] != None):
                numax = surface_effects['numax']
            else:
                numax = Model.numaxscl
            if (formula.lower() == 'lorentz3' ):
                dnu = numax*(par[2]  + par[0]/(1. + (nu/numax)**par[1]))
            elif ( (formula.lower() == 'lorentz2') | (formula.lower() == 'lorentz') ):
                dnu = par[0]*numax*(1./(1. + (nu/numax)**par[1]) - 1.)
            elif (formula.lower() == 'kb2008' ):
                dnu =  par[0]*numax*(nu/numax)**par[1]
            n = nu.size
            print "# surface effects"
            print ("# par0, par1 : %f, %f") % (par[0],par[1])
#            print "# nu, dnu, nu_corr "
#            for i in range(n):
#                print ("%f %f %f") % (nu[i],dnu[i],nu[i]+dnu[i])
            return dnu
    return np.zeros(nu.shape) 


def largesep(Model,Params,Target,Settings,path=''):
    name = Model.name
    modesset = Settings['modes']
    oscprog = modesset['oscprog']
    if oscprog == None:
        raise OSMError("Unspecified pulsation programme")
    if ( oscprog == 'adipls'):
        modes_ = read_agsm(path+name+'.gsm')
    elif ( oscprog == 'mad' ):
        modes_ = read_mad(path+name+'-nad.nad')
    else:
        print "%s : unknown oscillation programme" % oscprog
        return np.empty(0)

    modes = np.zeros( (modes_.shape[0],4))
    modes[:,0:3] = modes_[:,0:3]

    # adding surface effects
    dnuse =  surface_effects(modes[:,2],Model,Params,Settings)
    modes[:,2] = modes[:,2]  +dnuse

    if(modesset == None):
        (y,covar,yn) = dnu(modes)
        return y.mean()
    else:
        l = modesset['l']
        print 'we select the degrees l:',np.array_str(l)
        dn = modesset['dn']
        nmin = modesset['nmin'] 
        nmax = modesset['nmax'] 
        print ('we select the orders from nmin=%d to nmax=%d with a tolerrance dn = %d') % (nmin,nmax,dn)
        nm = modes.shape[0]
        lsep = np.zeros(dn*2+1)
        print '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
        for i in range(dn*2+1):
            shift = i - dn
            print ('shift = %d') %(shift)
            selected = []
            for li in l:
                nl = 0
                for k in range(nm):
                    for n in range(nmin+shift,nmax+shift+1):
                        if( (abs(modes[k,0] -  n) < 1e-5 ) & ( abs(modes[k,1] - li)  < 1e-5) ):
                            print ('selected modes (l,n,nu,w2) = (%d,%d,%f,%f)') % (li,n,modes[k,2],modes[k,3])
                            selected.append(k)
                            nl = nl + 1
                if(nl ==0):
                    print ('l=%d modes missing') % (li)
                else:
                    if(nl < 2):
                        print ('WARNING ! not enough l=%d modes') % (li)
            nms =len(selected) 
            print ('number of modes selected = %d') % (nms)
            if(nms > 1 ):
                (y,covar,yn) = dnu(modes[selected,:])
                lsep[i]  = y.mean()
            else:
                if(nms ==0):
                    print 'WARNING ! no modes matching the requirements'
                if(nms < 2):
                    print 'WARNING ! not enough modes matching the requirements'
                lsep[i] = 0.
            print ('largsep = %f') % (lsep[i])
            print '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
        i = np.argmin( abs(Target.value-lsep) )
        print ('closest value of largesep = %f (shift = %d)') % (lsep[i],i-dn)
        return lsep[i]



def seismic_model(Model,sc,Params,Settings,path=''):

    name = Model.name
    modesset = Settings['modes']
    oscprog = modesset['oscprog']
    if oscprog == None:
        raise OSMError("Unspecified pulsation program")
    if ( oscprog == 'adipls'):
        modes_ = read_agsm(path+name+'.gsm')
    elif ( oscprog == 'mad' ):
        modes_ = read_mad(path+name+'-nad.nad')
    else:
        print "%s : unknown oscillation program" % oscprog
        return np.empty(0)    
        
    modes = np.zeros( (modes_.shape[0],4))
    modes[:,0:3] = modes_[:,0:3]

    # adding surface effects
    dnuse = surface_effects(modes[:,2],Model,Params,Settings)
    modes[:,2] = modes[:,2] + dnuse

    modesn = np.empty( (0,4) )

    # we check that the th. frequencies cover the observations
    lval0 = np.unique(modes[:,1])
    for lc in sc.lval:
        found = False
        for l in lval0:
            if(lc == l):
                found = True
                j = (np.where(modes[:,1] == lc))[0]
                jc = (np.where(sc.l == lc))[0]
                LS = (sc.nu[jc[1:]] - sc.nu[jc[0:-1]] ).mean()
                if (sc.matching == 'continuous_frequency'):
                    #                print 'LS=',LS
                    if ( np.min(modes[j,2]) > np.min(sc.nu[jc])+ LS/2. or np.max(modes[j,2]) < np.max(sc.nu[jc]) - LS/2. ):
                        print 'warning in seismic_model: in  '+ name +', l=' +  ("%i") % (lc) + ' modes do not cover the observational range'

                    imin = np.argmin( np.abs(modes[j,2] - np.min(sc.nu[jc])) )
                elif (sc.matching == 'continuous_order'):
                    imin = np.argmin( np.abs(modes[j,0] - sc.n[jc][0]) )
                elif (sc.matching == 'frequency'):
                    count = 0
                    for k in j:
                        for i in jc:
                            if ( math.fabs(sc.nu[i]-modes[k,2])  < LS/2.  ):
                                modesn = np.append( modesn, [modes[k,:]],axis=0)
                                count += 1
                            
                    if (count <=0):
                        print ('warning in seismic_model: in %s, no matching between theoretical and observed l=' +  "%i modes") % (name,lc)
                elif (sc.matching == 'order'):
                    count = 0
                    for k in j:
                        for i in jc:
                            if ( math.fabs(sc.n[i]-modes[k,0])  < 1e-10  ):
                                modesn = np.append( modesn, [modes[k,:]], axis=0 )
                                count += 1
                    if (count <=0):
                        print ('warning in seismic_model: in %s, no matching between theoretical and observed l=' +  "%i modes") % (name,lc)
                    
                if ( (sc.matching == 'continuous_frequency') | (sc.matching == 'continuous_order') ):
                    print ('first l=%i modes: n=%i  nu=%g') % (lc,modes[j[imin],0],modes[j[imin],2])
                    imax = imin + len(jc) -1
                    if( len(jc) > len(j)-imin):
                        print 'error in seismic_model: in '+ name +'  insufficient number of  l=' +  ("%i") % (lc) + ' we need ' + ("%i") %(len(jc)) + ' while only ' + ("%i") %(len(j)-imin) + '  lie in the frequency range' 
                        return np.empty(0)            
                    else:
                        print ('last l=%i modes: n=%i  nu=%g') % (lc,modes[j[imax],0],modes[j[imax],2])
                        modesn = np.append( modesn , modes[j[imin]:j[imax]+1,:] ,axis=0)

        if(not found):
            print 'error in seismic_model: in '+ name +'  l=' +  ("%i") % (lc) + ' modes are missing'
            return np.empty(0)


    print 'selected th. frequencies:'
    print 'n   l   nu [muHz]'
    for i in  range(modesn.shape[0]):
        print ("%2i %2i %15.5f" )% (modesn[i,0],modesn[i,1],modesn[i,2])

    Y = np.empty((0))
    for type in sc.types:
        if( type == 'dnu'):
            (y,coef,yn) = dnu(modesn)
            Y = np.append(Y,y)
        elif( type == 'dnu0'):
            (y,coef,yn) = dnu(modesn,lval=[0])
            Y = np.append(Y,y)
        elif( type == 'dnu1'):
            (y,coef,yn) = dnu(modesn,lval=[1])
            Y = np.append(Y,y)
        elif( type == 'dnu2'):
            (y,coef,yn) = dnu(modesn,lval=[2])
            Y = np.append(Y,y)
        elif( type == 'd01'):
            (y,coef,yn) = d01(modesn)
            Y = np.append(Y,y)
        elif( type == 'd02'):
            (y,coef,yn) = d02(modesn)
            Y = np.append(Y,y)
        elif( type == 'rd02'):
            (y,coef,a,b,yn) = rd02(modesn)
            Y = np.append(Y,y)
        elif( type == 'sd'):
            (y,coef,yn) = sd(modesn)
            Y = np.append(Y,y)
        elif( type == 'sd01'):
            (y,coef,yn) = sd01(modesn)
            Y = np.append(Y,y)
        elif( type == 'sd10'):
            (y,coef,yn) = sd10(modesn)
            Y = np.append(Y,y)
        elif( type == 'rsd01'):
            (y,coef,a,b,yn) = rsd01(modesn)
            Y = np.append(Y,y)
        elif( type == 'rsd10'):
            (y,coef,a,b,yn) = rsd10(modesn)
            Y = np.append(Y,y)
        elif( type == 'nu'):
            Y = np.append(Y,modesn[:,2])
        else:
            print "error in seismic_model: seimic constraints of type '"+ type + "' is not handled"

            Y = np.append(Y,y)


    return Y
